# mwpot_helpers.py: auxiliary functions to help in setting up various
# Milky-Way-like potentials
# (for now, functions required to setup the Dehnen & Binney (1998), 
#  Binney & Tremaine (2008), and McMillan (2017) potentials)
import numpy
def expexp_dens(R,z,Rd,zd,Sigma0):
    """rho(R,z) = Sigma_0/(2zd) exp(-|z|/zd-R/Rd)"""
    return Sigma0/(2*zd)*numpy.exp(-numpy.fabs(z)/zd-R/Rd)

def expexp_dens_with_hole(R,z,Rd,Rm,zd,Sigma0):
    """rho(R,z) = Sigma0 / (4zd) exp(-Rm/R-R/Rd-|z|/zd)"""
    if R == 0.:
        return 0.
    return Sigma0/(2*zd)*numpy.exp(-Rm/R-R/Rd-numpy.fabs(z)/zd)

def expsech2_dens_with_hole(R,z,Rd,Rm,zd,Sigma0):
    """rho(R,z) = Sigma0 / (4zd) exp(-Rm/R-R/Rd)*sech(z/[2zd])^2"""
    if R == 0.:
        return 0.
    return Sigma0/(4*zd)*numpy.exp(-Rm/R-R/Rd)/numpy.cosh(z/(2*zd))**2

def core_pow_dens_with_cut(R,z,alpha,r0,rcut,rho0,q):
    """rho(R,z) = rho0(1+r'/r0)^-alpha exp(-[r'/rcut]^2)
    r' = sqrt(R^2+z^2/q^2"""
    rdash= numpy.sqrt(R**2+(z/q)**2)
    return rho0/(1+rdash/r0)**alpha*numpy.exp(-(rdash/rcut)**2)

def pow_dens_with_cut(R,z,alpha,r0,rcut,rho0,q):
    """rho(R,z) = rho0(1+r'/r0)^-alpha exp(-[r'/rcut]^2)
    r' = sqrt(R^2+z^2/q^2"""
    rdash= numpy.sqrt(R**2+(z/q)**2)
    return rho0/(rdash/r0)**alpha*numpy.exp(-(rdash/rcut)**2)



###############################################################################
#   WrapperPotential.py: Super-class for wrapper potentials
###############################################################################
from .Potential import Potential, _isNonAxi, _dim
from .planarPotential import planarPotential
from .Potential import evaluatePotentials, \
    evaluateRforces, evaluatephiforces, evaluatezforces, \
    evaluateR2derivs, evaluatez2derivs, \
    evaluateRzderivs, evaluateDensities
from .planarPotential import evaluateplanarPotentials, \
    evaluateplanarRforces, evaluateplanarphiforces, \
    evaluateplanarR2derivs

def _new_obj(cls, kwargs, args):
    """Maps kwargs to cls.__new__"""
    return cls.__new__(cls, *args, **kwargs)

class parentWrapperPotential(object):
    """'Dummy' class only used to delegate wrappers to either 2D planarWrapperPotential or 3D WrapperPotential based on pot's dimensionality, using a little python object creation magic..."""
    def __new__(cls,*args,**kwargs):
        if kwargs.pop('_init',False):
            # When we get here recursively, just create new object
            return object.__new__(cls)
        # Decide whether superclass is Wrapper or planarWrapper based on dim
        pot= kwargs.get('pot',None)
        if _dim(pot) == 2:
            parentWrapperPotential= planarWrapperPotential
        elif _dim(pot) == 3:
            parentWrapperPotential= WrapperPotential
        else:
            raise ValueError("WrapperPotentials are only supported in 3D and 2D")
        # Create object from custom class that derives from correct wrapper,
        # make sure to turn off normalization for all wrappers
        kwargs['_init']= True # to break recursion above
        # __reduce__ method to allow pickling
        reduce= lambda self: (_new_obj, (cls, kwargs, args), self.__dict__)
        out= type.__new__(type,'_%s' % cls.__name__,
                          (parentWrapperPotential,cls),
                          {'normalize':property(),
                           '__reduce__':reduce})(*args,**kwargs)
        kwargs.pop('_init',False)
        # This runs init for the subclass (the specific wrapper)
        cls.__init__(out,*args,**kwargs)
        return out

class WrapperPotential(Potential):
    def __init__(self,amp=1.,pot=None,ro=None,vo=None,_init=None,**kwargs):
        """
        NAME:

           __init__

        PURPOSE:

           initialize a WrapperPotential, a super-class for wrapper potentials

        INPUT:

           amp - amplitude to be applied to the potential (default: 1.)

           pot - Potential instance or list thereof; the amplitude of this will be grown by this wrapper

        OUTPUT:

           (none)

        HISTORY:

           2017-06-26 - Started - Bovy (UofT)

        """
        if not _init: return None # Don't run __init__ at the end of setup
        Potential.__init__(self,amp=amp,ro=ro,vo=vo)
        self._pot= pot
        self.isNonAxi= _isNonAxi(self._pot)

    def __getattr__(self,attribute):
        if attribute == '_evaluate' \
                or attribute == '_Rforce' or attribute == '_zforce' \
                or attribute == '_phiforce' \
                or attribute == '_R2deriv' or attribute == '_z2deriv' \
                or attribute == '_Rzderiv' or attribute == '_phi2deriv' \
                or attribute == '_Rphideriv' or attribute == '_dens':
            return lambda R,Z,phi=0.,t=0.: \
                self._wrap(attribute,R,Z,phi=phi,t=t)
        else:
            return super(WrapperPotential,self).__getattr__(attribute)

    def _wrap_pot_func(self,attribute):
        if attribute == '_evaluate':
            return evaluatePotentials
        elif attribute == '_dens':
            return evaluateDensities
        elif attribute == '_Rforce':
            return evaluateRforces
        elif attribute == '_zforce':
            return evaluatezforces
        elif attribute == '_phiforce':
            return evaluatephiforces
        elif attribute == '_R2deriv':
            return evaluateR2derivs
        elif attribute == '_z2deriv':
            return evaluatez2derivs
        elif attribute == '_Rzderiv':
            return evaluateRzderivs
        elif attribute == '_phi2deriv':
            return lambda p,R,Z,phi=0.,t=0.: \
                evaluatePotentials(p,R,Z,phi=phi,t=t,dphi=2)
        elif attribute == '_Rphideriv':
            return lambda p,R,Z,phi=0.,t=0.: \
                evaluatePotentials(p,R,Z,phi=phi,t=t,dR=1,dphi=1)
        else: #pragma: no cover
            raise AttributeError("Attribute %s not found in for this WrapperPotential" % attribute)

class planarWrapperPotential(planarPotential):
    def __init__(self,amp=1.,pot=None,ro=None,vo=None,_init=None,**kwargs):
        """
        NAME:

           __init__

        PURPOSE:

           initialize a WrapperPotential, a super-class for wrapper potentials

        INPUT:

           amp - amplitude to be applied to the potential (default: 1.)

           pot - Potential instance or list thereof; the amplitude of this will be grown by this wrapper

        OUTPUT:

           (none)

        HISTORY:

           2017-06-26 - Started - Bovy (UofT)

        """
        if not _init: return None # Don't run __init__ at the end of setup
        planarPotential.__init__(self,amp=amp,ro=ro,vo=vo)
        self._pot= pot
        self.isNonAxi= _isNonAxi(self._pot)

    def __getattr__(self,attribute):
        if attribute == '_evaluate' \
                or attribute == '_Rforce' \
                or attribute == '_phiforce' \
                or attribute == '_R2deriv' \
                or attribute == '_phi2deriv' \
                or attribute == '_Rphideriv':
            return lambda R,phi=0.,t=0.: \
                self._wrap(attribute,R,phi=phi,t=t)
        else:
            return super(planarWrapperPotential,self).__getattr__(attribute)

    def _wrap_pot_func(self,attribute):
        if attribute == '_evaluate':
            return evaluateplanarPotentials
        elif attribute == '_Rforce':
            return evaluateplanarRforces
        elif attribute == '_phiforce':
            return evaluateplanarphiforces
        elif attribute == '_R2deriv':
            return evaluateplanarR2derivs
        elif attribute == '_phi2deriv':
            return lambda p,R,phi=0.,t=0.: \
                evaluateplanarPotentials(p,R,phi=phi,t=t,dphi=2)
        elif attribute == '_Rphideriv':
            return lambda p,R,phi=0.,t=0.: \
                evaluateplanarPotentials(p,R,phi=phi,t=t,dR=1,dphi=1)
        else: #pragma: no cover
            raise AttributeError("Attribute %s not found in for this WrapperPotential" % attribute)




###############################################################################
#   FerrersPotential.py: General class for triaxial Ferrers Potential
#
#       rho(r) = amp/[a^3 b c pi^1.5] Gamma(n+5/2)/Gamma(n+1) (1 - (m/a)^2)^n
#
#       with
#
#       m^2 = x^2 + y^2/b^2 + z^2/c^2
########################################################################
import numpy as np
import hashlib
from scipy import integrate
from scipy.special import gamma
from galpy.util import bovy_conversion, bovy_coords
from .Potential import Potential, _APY_LOADED
if _APY_LOADED:
    from astropy import units

class FerrersPotential(Potential):
    """Class that implements triaxial Ferrers potential for the ellipsoidal density profile with the short axis along the z-direction

    .. math::

        \\rho(x,y,z) = \\frac{\\mathrm{amp}}{\\pi^{1.5} a^3 b c} \\frac{\\Gamma(n+\\frac{5}{2})}{\\Gamma(n+1)}\\,(1-(m/a)^2)^n

    with

    .. math::

        m^2 = x'^2 + \\frac{y'^2}{b^2}+\\frac{z'^2}{c^2}

    and :math:`(x',y',z')` is a rotated frame wrt :math:`(x,y,z)`
    so that the major axis is aligned with :math:`x'`.

    Note that this potential has not yet been optimized for speed and has no C implementation, so orbit integration is currently slow.
    """

    def __init__(self,amp=1.,a=1.,n=2,b=0.35,c=0.2375,omegab=0.,
                 pa=0.,normalize=False,ro=None,vo=None):
        """
        NAME:

           __init__

        PURPOSE:

           initialize a Ferrers potential

        INPUT:

           amp - total mass of the ellipsoid determines the amplitude of the potential; can be a Quantity with units of mass or Gxmass

           a - scale radius (can be Quantity)

           n - power of Ferrers density (n > 0)

           b - y-to-x axis ratio of the density

           c - z-to-x axis ratio of the density

           omegab - rotation speed of the ellipsoid (can be Quantity)

           pa= (None) If set, the position angle of the x axis (rad or Quantity)

           normalize - if True, normalize such that vc(1.,0.)=1., or, if given as a number, such that the force is this fraction of the force necessary to make vc(1.,0.)=1.

           ro=, vo= distance and velocity scales for translation into internal units (default from configuration file)

        OUTPUT:

           (none)

        """
        Potential.__init__(self,amp=amp,ro=ro,vo=vo,amp_units='mass')
        if _APY_LOADED and isinstance(a,units.Quantity):
            a= a.to(units.kpc).value/self._ro
        if _APY_LOADED and isinstance(omegab,units.Quantity):
            omegab= omegab.to(units.km/units.s/units.kpc).value\
                /bovy_conversion.freq_in_kmskpc(self._vo,self._ro)
        if _APY_LOADED and isinstance(pa,units.Quantity):
            pa= pa.to(units.rad).value
        self.a= a
        self._scale= self.a
        if n <= 0:
            raise ValueError('FerrersPotential requires n > 0')
        self.n= n
        self._b= b
        self._c= c
        self._omegab= omegab
        self._a2= self.a**2
        self._b2= self._b**2.
        self._c2= self._c**2.
        self._force_hash= None
        self._pa = pa
        self._rhoc_M = gamma(n+2.5)/gamma(n+1) / np.pi**1.5/a**3/b/c
        if normalize or \
                (isinstance(normalize,(int,float)) \
                     and not isinstance(normalize,bool)): #pragma: no cover
            self.normalize(normalize)
        if np.fabs(self._b-1.) > 10.**-10.:
            self.isNonAxi= True
        return None

    def _evaluate(self,R,z,phi=0.,t=0.):
        """
        NAME:
           _evaluate
        PURPOSE:
           evaluate the potential at R,z
        INPUT:
           R - Galactocentric cylindrical radius
           z - vertical height
           phi - azimuth
           t - time
        OUTPUT:
           Phi(R,z)
        """
        if not self.isNonAxi:
            phi= 0.
        x,y,z= bovy_coords.cyl_to_rect(R,phi,z)
        xy= np.dot(self.rot(t),np.array([x,y]))
        x,y= xy[0],xy[1]
        return self._evaluate_xyz(x,y,z)

    def _evaluate_xyz(self,x,y,z=0.):
        """Evaluation of the potential as a function of (x,y,z) in the 
        aligned coordinate frame"""
        return -np.pi * self._rhoc_M /(self.n+1.)*self.a**3*self._b*self._c * \
            _potInt(x, y, z, self._a2, self._b2*self._a2, self._c2*self._a2, self.n)

    def _Rforce(self,R,z,phi=0.,t=0.):
        """
        NAME:
           _Rforce
        PURPOSE:T
           evaluate the radial force for this potential
        INPUT:
           R - Galactocentric cylindrical radius
           z - vertical height
           phi - azimuth
           t - time
        OUTPUT:
           the radial force
        """
        if not self.isNonAxi:
            phi= 0.
        self._compute_xyzforces(R,z,phi,t)
        return np.cos(phi)*self._cached_Fx+np.sin(phi)*self._cached_Fy

    def _phiforce(self,R,z,phi=0.,t=0.):
        """
        NAME:
           _phiforce
        PURPOSE:
           evaluate the azimuthal force for this potential
        INPUT:
           R - Galactocentric cylindrical radius
           z - vertical height
           phi - azimuth
           t - time
        OUTPUT:
           the azimuthal force
        """
        if not self.isNonAxi:
            phi= 0.
        self._compute_xyzforces(R,z,phi,t)
        return R*(-np.sin(phi)*self._cached_Fx\
                       +np.cos(phi)*self._cached_Fy)

    def _zforce(self,R,z,phi=0.,t=0.):
        """
        NAME:
           _zforce
        PURPOSE:
           evaluate the vertical force for this potential
        INPUT:
           R - Galactocentric cylindrical radius
           z - vertical height
           phi - azimuth
           t - time
        OUTPUT:
           the vertical force
        """
        if not self.isNonAxi:
            phi= 0.
        self._compute_xyzforces(R,z,phi,t)
        return self._cached_Fz

    def _compute_xyz(self,R,phi,z,t):
        return bovy_coords.cyl_to_rect(R,phi-self._pa-self._omegab*t,z)

    def _compute_xyzforces(self,R,z,phi,t):
        # Compute all rectangular forces
        new_hash= hashlib.md5(np.array([R,phi,z,t])).hexdigest()
        if new_hash == self._force_hash:
            Fx= self._cached_Fx
            Fy= self._cached_Fy
            Fz= self._cached_Fz
        else:
            x,y,z= self._compute_xyz(R,phi,z,t)
            Fx= self._xforce_xyz(x,y,z)
            Fy= self._yforce_xyz(x,y,z)
            Fz= self._zforce_xyz(x,y,z)
            self._force_hash= new_hash
            tp= self._pa+self._omegab*t
            cp, sp= np.cos(tp), np.sin(tp)
            self._cached_Fx= cp*Fx-sp*Fy
            self._cached_Fy= sp*Fx+cp*Fy
            self._cached_Fz= Fz

    def _xforce_xyz(self,x,y,z):
        """Evaluation of the x force as a function of (x,y,z) in the aligned
        coordinate frame"""
        return -2.*np.pi*self._rhoc_M * self.a**3*self._b*self._c * \
            _forceInt(x, y, z, self._a2, self._b2*self._a2, self._c2*self._a2, self.n, 0)

    def _yforce_xyz(self,x,y,z):
        """Evaluation of the y force as a function of (x,y,z) in the aligned
        coordinate frame"""
        return -2.*np.pi*self._rhoc_M * self.a**3*self._b*self._c * \
            _forceInt(x, y, z, self._a2, self._b2*self._a2, self._c2*self._a2, self.n, 1)

    def _zforce_xyz(self,x,y,z):
        """Evaluation of the z force as a function of (x,y,z) in the aligned
        coordinate frame"""
        return -2.*np.pi*self._rhoc_M * self.a**3*self._b*self._c * \
            _forceInt(x, y, z, self._a2, self._b2*self._a2, self._c2*self._a2, self.n, 2)

    def _R2deriv(self,R,z,phi=0.,t=0.):
        """
        NAME:
           _R2deriv
        PURPOSE:
           evaluate the second radial derivative for this potential
        INPUT:
           R - Galactocentric cylindrical radius
           z - vertical height
           phi - azimuth
           t - time
        OUTPUT:
           the second radial derivative
        """
        if not self.isNonAxi:
            phi= 0.
        x,y,z= self._compute_xyz(R,phi,z,t)
        phixxa= self._2ndderiv_xyz(x,y,z,0,0)
        phixya= self._2ndderiv_xyz(x,y,z,0,1)
        phiyya= self._2ndderiv_xyz(x,y,z,1,1)
        ang = self._omegab*t + self._pa
        c, s = np.cos(ang), np.sin(ang)
        phixx = c**2*phixxa + 2.*c*s*phixya + s**2*phiyya
        phixy = (c**2-s**2)*phixya + c*s*(phiyya - phixxa)
        phiyy = s**2*phixxa - 2.*c*s*phixya + c**2*phiyya
        return np.cos(phi)**2.*phixx + np.sin(phi)**2.*phiyy + \
            2.*np.cos(phi)*np.sin(phi)*phixy

    def _Rzderiv(self,R,z,phi=0.,t=0.):
        """
        NAME:
           _Rzderiv
        PURPOSE:
           evaluate the mixed radial, vertical derivative for this potential
        INPUT:
           R - Galactocentric cylindrical radius
           z - vertical height
           phi - azimuth
           t - time
        OUTPUT:
           the mixed radial, vertical derivative
        """
        if not self.isNonAxi:
            phi= 0.
        x,y,z= self._compute_xyz(R,phi,z,t)
        phixza= self._2ndderiv_xyz(x,y,z,0,2)
        phiyza= self._2ndderiv_xyz(x,y,z,1,2)
        ang = self._omegab*t + self._pa
        c, s = np.cos(ang), np.sin(ang)
        phixz = c*phixza + s*phiyza
        phiyz = -s*phixza + c*phiyza
        return np.cos(phi)*phixz + np.sin(phi)*phiyz

    def _z2deriv(self,R,z,phi=0.,t=0.):
        """
        NAME:
           _z2deriv
        PURPOSE:
           evaluate the second vertical derivative for this potential
        INPUT:
           R - Galactocentric cylindrical radius
           z - vertical height
           phi - azimuth
           t - time
        OUTPUT:
           the second vertical derivative
        """
        if not self.isNonAxi:
            phi= 0.
        x,y,z= self._compute_xyz(R,phi,z,t)
        return self._2ndderiv_xyz(x,y,z,2,2)

    def _phi2deriv(self,R,z,phi=0.,t=0.):
        """
        NAME:
           _phi2deriv
        PURPOSE:
           evaluate the second azimuthal derivative for this potential
        INPUT:
           R - Galactocentric cylindrical radius
           z - vertical height
           phi - azimuth
           t - time
        OUTPUT:
           the second azimuthal derivative
        """
        if not self.isNonAxi:
            phi= 0.
        x,y,z= self._compute_xyz(R,phi,z,t)
        Fx= self._xforce_xyz(x,y,z)
        Fy= self._yforce_xyz(x,y,z)
        Fxy= np.dot(self.rot(t, transposed = True),np.array([Fx,Fy]))
        Fx, Fy= Fxy[0], Fxy[1]
        phixxa= self._2ndderiv_xyz(x,y,z,0,0)
        phixya= self._2ndderiv_xyz(x,y,z,0,1)
        phiyya= self._2ndderiv_xyz(x,y,z,1,1)
        ang = self._omegab*t + self._pa
        c, s = np.cos(ang), np.sin(ang)
        phixx = c**2*phixxa + 2.*c*s*phixya + s**2*phiyya
        phixy = (c**2-s**2)*phixya + c*s*(phiyya - phixxa)
        phiyy = s**2*phixxa - 2.*c*s*phixya + c**2*phiyya
        return R**2.*(np.sin(phi)**2.*phixx+np.cos(phi)**2.*phiyy\
                          -2.*np.cos(phi)*np.sin(phi)*phixy)\
                          +R*(np.cos(phi)*Fx+np.sin(phi)*Fy)

    def _Rphideriv(self,R,z,phi=0.,t=0.):
        """
        NAME:
           _Rphideriv
        PURPOSE:
           evaluate the mixed radial, azimuthal derivative for this potential
        INPUT:
           R - Galactocentric cylindrical radius
           z - vertical height
           phi - azimuth
           t - time
        OUTPUT:
           the mixed radial, azimuthal derivative
        """
        if not self.isNonAxi:
            phi= 0.
        x,y,z= self._compute_xyz(R,phi,z,t)
        Fx= self._xforce_xyz(x,y,z)
        Fy= self._yforce_xyz(x,y,z)
        Fxy= np.dot(self.rot(t, transposed = True),np.array([Fx,Fy]))
        Fx, Fy= Fxy[0], Fxy[1]
        phixxa= self._2ndderiv_xyz(x,y,z,0,0)
        phixya= self._2ndderiv_xyz(x,y,z,0,1)
        phiyya= self._2ndderiv_xyz(x,y,z,1,1)
        ang = self._omegab*t + self._pa
        c, s = np.cos(ang), np.sin(ang)
        phixx = c**2*phixxa + 2.*c*s*phixya + s**2*phiyya
        phixy = (c**2-s**2)*phixya + c*s*(phiyya - phixxa)
        phiyy = s**2*phixxa - 2.*c*s*phixya + c**2*phiyya
        return R*np.cos(phi)*np.sin(phi)*\
            (phiyy-phixx)+R*np.cos(2.*(phi))*phixy\
            +np.sin(phi)*Fx-np.cos(phi)*Fy

    def _2ndderiv_xyz(self,x,y,z,i,j):
        """General 2nd derivative of the potential as a function of (x,y,z)
        in the aligned coordinate frame, d^2\Phi/dx_i/dx_j"""
        return -np.pi*self._rhoc_M*self.a**3*self._b*self._c *\
            _2ndDerivInt(x,y,z,self._a2,self._b2*self._a2,self._c2*self._a2,self.n,i,j)

    def _dens(self,R,z,phi=0.,t=0.):
        """
        NAME:
           _dens
        PURPOSE:
           evaluate the density for this potential
        INPUT:
           R - Galactocentric cylindrical radius
           z - vertical height
           phi - azimuth
           t - time
        OUTPUT:
           the density
        """
        x,y,z= self._compute_xyz(R,phi,z,t)
        m2 = x**2/self._a2+y**2/self._b2+z**2/self._c2
        if m2 < 1:
            return self._rhoc_M*(1.-m2/self.a**2)**self.n
        else:
            return 0.

    def OmegaP(self):
        """
        NAME:
           OmegaP
        PURPOSE:
           return the pattern speed
        INPUT:
           (none)
        OUTPUT:
           pattern speed
        """
        return self._omegab

    def rot(self, t=0., transposed=False):
        """2D Rotation matrix for non-zero pa or pattern speed
           to goto the aligned coordinates
        """
        rotmat = np.array(
            [[np.cos(self._pa+self._omegab*t),np.sin(self._pa+self._omegab*t)],
             [-np.sin(self._pa+self._omegab*t),np.cos(self._pa+self._omegab*t)]])
        if transposed:
            return rotmat.T
        else:
            return rotmat

def _potInt(x,y,z,a2,b2,c2,n):
    """Integral involed in the potential at (x,y,z)
    integrates 1/A B^(n+1) where
    A = sqrt((tau+a)(tau+b)(tau+c)) and B = (1-x^2/(tau+a)-y^2/(tau+b)-z^2/(tau+c))
    from lambda to infty with respect to tau.
    The lower limit lambda is given by lowerlim function.
    """
    def integrand(tau):
        return _FracInt(x, y, z, a2, b2, c2, tau, n+1)
    return integrate.quad(integrand, lowerlim(x**2,y**2,z**2,a2,b2,c2), np.inf)[0]

def _forceInt(x,y,z,a2,b2,c2,n,i):
    """Integral involved in the force at (x,y,z)
    integrates 1/A B^n (x_i/(tau+a_i)) where
    A = sqrt((tau+a)(tau+b)(tau+c)) and B = (1-x^2/(tau+a)-y^2/(tau+b)-z^2/(tau+c))
    from lambda to infty with respect to tau.
    The lower limit lambda is given by lowerlim function.
    """
    def integrand(tau):
        return (x*(i==0) + y*(i==1) + z*(i==2))/(a2*(i==0) + b2*(i==1) + c2*(i==2) + tau) * \
            _FracInt(x, y, z, a2, b2, c2, tau, n)
    return integrate.quad(integrand, lowerlim(x**2, y**2, z**2, a2, b2, c2), np.inf, epsabs=1e-12)[0]

def _2ndDerivInt(x,y,z,a2,b2,c2,n,i,j):
    """Integral involved in second derivatives d^\Phi/(dx_i dx_j)
    integrate
        1/A B^(n-1) (-2 x_i/(tau+a_i)) (-2 x_j/(tau+a_j))
    when i /= j or
        1/A [ B^(n-1) 4n x_i^2 / (a_i+t)^2 + B^n -(-2/(a_i+t)) ]
    when i == j where
    A = sqrt((tau+a)(tau+b)(tau+c)) and B = (1-x^2/(tau+a)-y^2/(tau+b)-z^2/(tau+c))
    from lambda to infty with respect to tau
    The lower limit lambda is given by lowerlim function.
    This is a second derivative of _potInt.
    """
    def integrand(tau):
        if i!=j:
            return _FracInt(x,y,z,a2,b2,c2,tau,n-1)*n*\
                (1.+(-1.-2.*x/(tau+a2))*(i==0 or j==0))*\
                (1.+(-1.-2.*y/(tau+b2))*(i==1 or j==1))*\
                (1.+(-1.-2.*z/(tau+c2))*(i==2 or j==2))
        else:
            var2 = x**2*(i==0) + y**2*(i==1) + z**2*(i==2)
            coef2 = a2*(i==0) + b2*(i==1) + c2*(i==2)
            return _FracInt(x,y,z,a2,b2,c2,tau,n-1)*n*(4.*var2)/(tau+coef2)**2 +\
                _FracInt(x,y,z,a2,b2,c2,tau,n)*(-2./(tau+coef2))
    return integrate.quad(integrand, lowerlim(x**2,y**2,z**2,a2,b2,c2), np.inf)[0]

def _FracInt(x,y,z,a,b,c,tau,n):
    """Returns
                1                     x^2       y^2       z^2
    -------------------------- (1 - ------- - ------- - -------)^n
    sqrt(tau+a)(tau+b)(tau+c))       tau+a     tau+b     tau+c
    """
    denom = np.sqrt((a + tau)*(b + tau)*(c + tau))
    return (1. - x**2/(a + tau) - y**2/(b + tau) - z**2/(c + tau))**n / denom

def lowerlim(x, y, z, a, b, c):
    """Returns the real positive root of
      x/(a+t) + y/(b+t) + z/(c+t) = 1
    when x/a + y/b + z/c > 1 else zero
    """
    if x/a + y/b + z/c > 1:
        B = a + b + c - x - y - z
        C = a*b + a*c + b*c - a*y - a*z - b*x - b*z - c*x - c*y
        D = a*b*c - a*b*z - a*c*y - b*c*x
        r = np.roots([1,B,C,D])
        ll = r[~np.iscomplex(r) & (r>0.)]
        return ll[0].real
    else:
        return 0.



# The DF of a gap in a tidal stream
from functools import wraps
import copy
import numpy
import warnings
import multiprocessing
from scipy import integrate, interpolate, special
from galpy.util import galpyWarning, bovy_coords, multi, bovy_conversion
from galpy.util import _rotate_to_arbitrary_vector
from galpy.orbit import Orbit
from galpy.potential import evaluateRforces, MovingObjectPotential, \
    PlummerPotential
from .df import df, _APY_LOADED
from galpy.util.bovy_conversion import physical_conversion
from . import streamdf
from .streamdf import _determine_stream_track_single
from galpy.potential import flatten as flatten_potential
if _APY_LOADED:
    from astropy import units
def impact_check_range(func):
    """Decorator to check the range of interpolated kicks"""
    @wraps(func)
    def impact_wrapper(*args,**kwargs):
        if isinstance(args[1],numpy.ndarray):
            out= numpy.zeros(len(args[1]))
            goodIndx= (args[1] < args[0]._deltaAngleTrackImpact)*(args[1] > 0.)
            out[goodIndx]= func(args[0],args[1][goodIndx])
            return out
        elif args[1] >= args[0]._deltaAngleTrackImpact or args[1] <= 0.:
            return 0.
        else:
            return func(*args,**kwargs)
    return impact_wrapper
class streamgapdf(streamdf.streamdf):
    """The DF of a gap in a tidal stream"""
    def __init__(self,*args,**kwargs):
        """
        NAME:

           __init__

        PURPOSE:

           Initialize the DF of a gap in a stellar stream

        INPUT:

           streamdf args and kwargs

           Subhalo and impact parameters:

              impactb= impact parameter (can be Quantity)

              subhalovel= velocity of the subhalo shape=(3) (can be Quantity)

              timpact time since impact (can be Quantity)

              impact_angle= angle offset from progenitor at which the impact occurred (rad) (can be Quantity)

              Subhalo: specify either 1( mass and size of Plummer sphere or 2( general spherical-potential object (kick is numerically computed)

                 1( GM= mass of the subhalo (can be Quantity)

                    rs= size parameter of the subhalo (can be Quantity)

                 2( subhalopot= galpy potential object or list thereof (should be spherical)

                 3( hernquist= (False) if True, use Hernquist kicks for GM/rs

           deltaAngleTrackImpact= (None) angle to estimate the stream track over to determine the effect of the impact [similar to deltaAngleTrack] (rad)

           nTrackChunksImpact= (floor(deltaAngleTrack/0.15)+1) number of chunks to divide the progenitor track in near the impact [similar to nTrackChunks]

           nKickPoints= (30xnTrackChunksImpact) number of points along the stream to compute the kicks at (kicks are then interpolated); '30' chosen such that higherorderTrack can be set to False and get calculations accurate to > 99%

           nokicksetup= (False) if True, only run as far as setting up the coordinate transformation at the time of impact (useful when using this in streampepperdf)

           spline_order= (3) order of the spline to interpolate the kicks with

           higherorderTrack= (False) if True, calculate the track using higher-order terms

        OUTPUT:

           object

        HISTORY:

           2015-06-02 - Started - Bovy (IAS)

        """
        df.__init__(self,ro=kwargs.get('ro',None),vo=kwargs.get('vo',None))
        # Parse kwargs
        impactb= kwargs.pop('impactb',1.)
        if _APY_LOADED and isinstance(impactb,units.Quantity):
            impactb= impactb.to(units.kpc).value/self._ro
        subhalovel= kwargs.pop('subhalovel',numpy.array([0.,1.,0.]))
        if _APY_LOADED and isinstance(subhalovel,units.Quantity):
            subhalovel= subhalovel.to(units.km/units.s).value/self._vo
        hernquist= kwargs.pop('hernquist',False)
        GM= kwargs.pop('GM',None)
        if not GM is None \
                and _APY_LOADED and isinstance(GM,units.Quantity):
            # GM can be GM or M
            try:
                GM= GM.to(units.pc*units.km**2/units.s**2)\
                    .value\
                    /bovy_conversion.mass_in_msol(self._vo,self._ro)\
                    /bovy_conversion._G
            except units.UnitConversionError: pass
            GM= GM.to(units.Msun).value\
                /bovy_conversion.mass_in_msol(self._vo,self._ro)
        rs= kwargs.pop('rs',None)
        if not rs is None \
                and _APY_LOADED and isinstance(rs,units.Quantity):
            rs= rs.to(units.kpc).value/self._ro
        subhalopot= kwargs.pop('subhalopot',None)
        timpact= kwargs.pop('timpact',1.)
        if _APY_LOADED and isinstance(timpact,units.Quantity):
            timpact= timpact.to(units.Gyr).value\
                /bovy_conversion.time_in_Gyr(self._vo,self._ro)
        impact_angle= kwargs.pop('impact_angle',1.)
        if _APY_LOADED and isinstance(impact_angle,units.Quantity):
            impact_angle= impact_angle.to(units.rad).value
        nokicksetup= kwargs.pop('nokicksetup',False)
        deltaAngleTrackImpact= kwargs.pop('deltaAngleTrackImpact',None)
        nTrackChunksImpact= kwargs.pop('nTrackChunksImpact',None)
        nKickPoints= kwargs.pop('nKickPoints',None)
        spline_order= kwargs.pop('spline_order',3)
        higherorderTrack= kwargs.pop('higherorderTrack',False)
        # For setup later
        nTrackChunks= kwargs.pop('nTrackChunks',None)
        interpTrack= kwargs.pop('interpTrack',
                                streamdf._INTERPDURINGSETUP)
        useInterp= kwargs.pop('useInterp',
                              streamdf._USEINTERP)
        # Analytical Plummer or general potential?
        self._general_kick= GM is None or rs is None
        if self._general_kick and subhalopot is None:
            raise IOError("One of (GM=, rs=) or subhalopot= needs to be set to specify the subhalo's structure")
        # Now run the regular streamdf setup, but without calculating the
        # stream track (nosetup=True)
        kwargs['nosetup']= True
        super(streamgapdf,self).__init__(*args,**kwargs)
        # Setup the machinery to go between (x,v) and (Omega,theta)
        # near the impact
        self._determine_nTrackIterations(kwargs.get('nTrackIterations',None))
        self._determine_deltaAngleTrackImpact(deltaAngleTrackImpact,timpact)
        self._determine_impact_coordtransform(self._deltaAngleTrackImpact,
                                              nTrackChunksImpact,
                                              timpact,impact_angle)
        # Set nKickPoints
        if nKickPoints is None:
            self._nKickPoints= 30*self._nTrackChunksImpact
        else:
            self._nKickPoints= nKickPoints
        if nokicksetup: return None
        # Compute \Delta Omega ( \Delta \theta_perp) and \Delta theta,
        # setup interpolating function
        self._determine_deltav_kick(impact_angle,impactb,subhalovel,
                                    GM,rs,subhalopot,
                                    spline_order,hernquist)
        self._determine_deltaOmegaTheta_kick(spline_order)
        # Then pass everything to the normal streamdf setup
        self.nInterpolatedTrackChunks= 201 #more expensive now
        self._higherorderTrack= higherorderTrack
        super(streamgapdf,self)._determine_stream_track(nTrackChunks)
        self._useInterp= useInterp
        if interpTrack or self._useInterp:
            super(streamgapdf,self)._interpolate_stream_track()
            super(streamgapdf,self)._interpolate_stream_track_aA()
        super(streamgapdf,self).calc_stream_lb()
        return None

    def pOparapar(self,Opar,apar):
        """
        NAME:

           pOparapar

        PURPOSE:

           return the probability of a given parallel (frequency,angle) offset pair

        INPUT:

           Opar - parallel frequency offset (array) (can be Quantity)

           apar - parallel angle offset along the stream (scalar) (can be Quantity)

        OUTPUT:

           p(Opar,apar)

        HISTORY:

           2015-11-17 - Written - Bovy (UofT)

        """
        if _APY_LOADED and isinstance(Opar,units.Quantity):
            Opar= Opar.to(1/units.Gyr).value\
                /bovy_conversion.freq_in_Gyr(self._vo,self._ro)
        if _APY_LOADED and isinstance(apar,units.Quantity):
            apar= apar.to(units.rad).value
        if isinstance(Opar,(int,float,numpy.float32,numpy.float64)):
            Opar= numpy.array([Opar])
        out= numpy.zeros(len(Opar))
        # Compute ts and where they were at impact for all
        ts= apar/Opar
        apar_impact= apar-Opar*self._timpact
        dOpar_impact= self._kick_interpdOpar(apar_impact)
        Opar_b4impact= Opar-dOpar_impact
        # Evaluate the smooth model in the two regimes:
        # stripped before or after impact
        afterIndx= (ts < self._timpact)*(ts >= 0.)
        out[afterIndx]=\
            super(streamgapdf,self).pOparapar(Opar[afterIndx],
                                              apar)
        out[True^afterIndx]=\
            super(streamgapdf,self).pOparapar(Opar_b4impact[True^afterIndx],
                                              apar_impact[True^afterIndx],
                                              tdisrupt=
                                              self._tdisrupt-self._timpact)
        return out

    def _density_par(self,dangle,tdisrupt=None,approx=True,
                     higherorder=None):
        """The raw density as a function of parallel angle,
        approx= use faster method that directly integrates the spline
        representation"""
        if higherorder is None: higherorder= self._higherorderTrack
        if tdisrupt is None: tdisrupt= self._tdisrupt
        if approx:
            return self._density_par_approx(dangle,tdisrupt,
                                            higherorder=higherorder)
        else:
            return integrate.quad(lambda T: numpy.sqrt(self._sortedSigOEig[2])\
                                      *(1+T*T)/(1-T*T)**2.\
                                      *self.pOparapar(T/(1-T*T)\
                                                          *numpy.sqrt(self._sortedSigOEig[2])\
                                                          +self._meandO,dangle),
                                  -1.,1.)[0]

    def _density_par_approx(self,dangle,tdisrupt,_return_array=False,
                            higherorder=False):
        """Compute the density as a function of parallel angle using the 
        spline representation + approximations"""
        # First construct the breakpoints for this dangle
        Oparb= (dangle-self._kick_interpdOpar_poly.x)/self._timpact
        # Find the lower limit of the integration in the pw-linear-kick approx.
        lowbindx,lowx= self.minOpar(dangle,tdisrupt,_return_raw=True)
        lowbindx= numpy.arange(len(Oparb)-1)[lowbindx]
        Oparb[lowbindx+1]= Oparb[lowbindx]-lowx
        # Now integrate between breakpoints
        out= (0.5/(1.+self._kick_interpdOpar_poly.c[-2]*self._timpact)\
                           *(special.erf(1./numpy.sqrt(2.*self._sortedSigOEig[2])\
                                             *(Oparb[:-1]-self._kick_interpdOpar_poly.c[-1]-self._meandO))\
                                 -special.erf(1./numpy.sqrt(2.*self._sortedSigOEig[2])*(numpy.roll(Oparb,-1)[:-1]-self._kick_interpdOpar_poly.c[-1]-self._meandO\
                                                                                             -self._kick_interpdOpar_poly.c[-2]*self._timpact*(Oparb-numpy.roll(Oparb,-1))[:-1]))))
        if _return_array:
            return out
        out= numpy.sum(out[:lowbindx+1])
        if higherorder:
            # Add higher-order contribution
            out+= self._density_par_approx_higherorder(Oparb,lowbindx)
        # Add integration to infinity
        out+= 0.5*(1.+special.erf((self._meandO-Oparb[0])\
                                  /numpy.sqrt(2.*self._sortedSigOEig[2])))
        return out

    def _density_par_approx_higherorder(self,Oparb,lowbindx,
                                        _return_array=False,
                                        gaussxpolyInt=None):
        """Contribution from non-linear spline terms"""
        spline_order= self._kick_interpdOpar_raw._eval_args[2]
        if spline_order == 1: return 0.
        # Form all Gaussian-like integrals necessary
        ll= (numpy.roll(Oparb,-1)[:-1]-self._kick_interpdOpar_poly.c[-1]\
            -self._meandO\
            -self._kick_interpdOpar_poly.c[-2]*self._timpact\
            *(Oparb-numpy.roll(Oparb,-1))[:-1])\
            /numpy.sqrt(2.*self._sortedSigOEig[2])
        ul= (Oparb[:-1]-self._kick_interpdOpar_poly.c[-1]-self._meandO)\
            /numpy.sqrt(2.*self._sortedSigOEig[2])
        if gaussxpolyInt is None:
            gaussxpolyInt=\
                self._densMoments_approx_higherorder_gaussxpolyInts(\
                ll,ul,spline_order+1)
        # Now multiply in the coefficients for each order
        powers= numpy.tile(numpy.arange(spline_order+1)[::-1],
                           (len(ul),1)).T
        gaussxpolyInt*= -0.5*(-numpy.sqrt(2.))**(powers+1)\
            *self._sortedSigOEig[2]**(0.5*(powers-1))
        powers= numpy.tile(numpy.arange(spline_order+1)[::-1][:-2],
                           (len(ul),1)).T
        for jj in range(spline_order+1):
            gaussxpolyInt[-jj-1]*= numpy.sum(\
                self._kick_interpdOpar_poly.c[:-2]
                *self._timpact**powers
                /(1.+self._kick_interpdOpar_poly.c[-2]*self._timpact)
                **(powers+1)
                *special.binom(powers,jj)
                *(Oparb[:-1]-self._kick_interpdOpar_poly.c[-1]-self._meandO)
                **(powers-jj),axis=0)
        if _return_array:
            return numpy.sum(gaussxpolyInt,axis=0)
        else:
            return numpy.sum(gaussxpolyInt[:,:lowbindx+1])

    def _densMoments_approx_higherorder_gaussxpolyInts(self,ll,ul,maxj):
        """Calculate all of the polynomial x Gaussian integrals occuring 
        in the higher-order terms, recursively"""
        gaussxpolyInt= numpy.zeros((maxj,len(ul)))
        gaussxpolyInt[-1]= 1./numpy.sqrt(numpy.pi)\
            *(numpy.exp(-ll**2.)-numpy.exp(-ul**2.))
        gaussxpolyInt[-2]= 1./numpy.sqrt(numpy.pi)\
            *(numpy.exp(-ll**2.)*ll-numpy.exp(-ul**2.)*ul)\
            +0.5*(special.erf(ul)-special.erf(ll))
        for jj in range(maxj-2):
            gaussxpolyInt[-jj-3]= 1./numpy.sqrt(numpy.pi)\
            *(numpy.exp(-ll**2.)*ll**(jj+2)-numpy.exp(-ul**2.)*ul**(jj+2))\
            +0.5*(jj+2)*gaussxpolyInt[-jj-1]
        return gaussxpolyInt

    def minOpar(self,dangle,tdisrupt=None,_return_raw=False):
        """
        NAME:

           minOpar

        PURPOSE:

           return the approximate minimum parallel frequency at a given angle

        INPUT:

           dangle - parallel angle

        OUTPUT:

           minimum frequency that gets to this parallel angle

        HISTORY:

           2015-12-28 - Written - Bovy (UofT)

        """
        if tdisrupt is None: tdisrupt= self._tdisrupt
        # First construct the breakpoints for this dangle
        Oparb= (dangle-self._kick_interpdOpar_poly.x[:-1])/self._timpact
        # Find the lower limit of the integration in the pw-linear-kick approx.
        lowx= ((Oparb-self._kick_interpdOpar_poly.c[-1])\
                   *(tdisrupt-self._timpact)+Oparb*self._timpact-dangle)\
                   /((tdisrupt-self._timpact)\
                         *(1.+self._kick_interpdOpar_poly.c[-2]*self._timpact)\
                         +self._timpact)
        lowx[lowx < 0.]= numpy.inf
        lowbindx= numpy.argmin(lowx)
        if _return_raw:
            return (lowbindx,lowx[lowbindx])
        else:
            return Oparb[lowbindx]-lowx[lowbindx]

    @physical_conversion('frequency',pop=True)
    def meanOmega(self,dangle,oned=False,tdisrupt=None,approx=True,
                  higherorder=None):
        """
        NAME:

           meanOmega

        PURPOSE:

           calculate the mean frequency as a function of angle, assuming a uniform time distribution up to a maximum time

        INPUT:

           dangle - angle offset

           oned= (False) if True, return the 1D offset from the progenitor (along the direction of disruption)

           approx= (True) if True, compute the mean Omega by direct integration of the spline representation

           higherorder= (object-wide default higherorderTrack) if True, include higher-order spline terms in the approximate computation

        OUTPUT:

           mean Omega

        HISTORY:

           2015-11-17 - Written - Bovy (UofT)

        """
        if higherorder is None: higherorder= self._higherorderTrack
        if tdisrupt is None: tdisrupt= self._tdisrupt
        if approx:
            num= self._meanOmega_num_approx(dangle,tdisrupt,
                                            higherorder=higherorder)
        else:
            num=\
                integrate.quad(lambda T: (T/(1-T*T)\
                                              *numpy.sqrt(self._sortedSigOEig[2])\
                                              +self._meandO)\
                                   *numpy.sqrt(self._sortedSigOEig[2])\
                                   *(1+T*T)/(1-T*T)**2.\
                                   *self.pOparapar(T/(1-T*T)\
                                                       *numpy.sqrt(self._sortedSigOEig[2])\
                                                       +self._meandO,dangle),
                               -1.,1.)[0]
        denom= self._density_par(dangle,tdisrupt=tdisrupt,approx=approx,
                                 higherorder=higherorder)
        dO1D= num/denom
        if oned: return dO1D
        else:
            return self._progenitor_Omega+dO1D*self._dsigomeanProgDirection\
                *self._sigMeanSign

    def _meanOmega_num_approx(self,dangle,tdisrupt,higherorder=False):
        """Compute the numerator going into meanOmega using the direct integration of the spline representation"""
        # First construct the breakpoints for this dangle
        Oparb= (dangle-self._kick_interpdOpar_poly.x)/self._timpact
        # Find the lower limit of the integration in the pw-linear-kick approx.
        lowbindx,lowx= self.minOpar(dangle,tdisrupt,_return_raw=True)
        lowbindx= numpy.arange(len(Oparb)-1)[lowbindx]
        Oparb[lowbindx+1]= Oparb[lowbindx]-lowx
        # Now integrate between breakpoints
        out= numpy.sum(((Oparb[:-1]
                         +(self._meandO+self._kick_interpdOpar_poly.c[-1]
                           -Oparb[:-1])/
                         (1.+self._kick_interpdOpar_poly.c[-2]*self._timpact))
                        *self._density_par_approx(dangle,tdisrupt,
                                                  _return_array=True)
                        +numpy.sqrt(self._sortedSigOEig[2]/2./numpy.pi)/
                        (1.+self._kick_interpdOpar_poly.c[-2]*self._timpact)**2.
                        *(numpy.exp(-0.5*(Oparb[:-1]
                                      -self._kick_interpdOpar_poly.c[-1]
                                      -(1.+self._kick_interpdOpar_poly.c[-2]*self._timpact)
                                      *(Oparb-numpy.roll(Oparb,-1))[:-1]
                                      -self._meandO)**2.
                                     /self._sortedSigOEig[2])
                          -numpy.exp(-0.5*(Oparb[:-1]-self._kick_interpdOpar_poly.c[-1]
                                       -self._meandO)**2.
                                      /self._sortedSigOEig[2])))[:lowbindx+1])
        if higherorder:
            # Add higher-order contribution
            out+= self._meanOmega_num_approx_higherorder(Oparb,lowbindx)
        # Add integration to infinity
        out+= 0.5*(numpy.sqrt(2./numpy.pi)*numpy.sqrt(self._sortedSigOEig[2])\
                        *numpy.exp(-0.5*(self._meandO-Oparb[0])**2.\
                                        /self._sortedSigOEig[2])
                   +self._meandO
                   *(1.+special.erf((self._meandO-Oparb[0])
                                    /numpy.sqrt(2.*self._sortedSigOEig[2]))))
        return out

    def _meanOmega_num_approx_higherorder(self,Oparb,lowbindx):
        """Contribution from non-linear spline terms"""
        spline_order= self._kick_interpdOpar_raw._eval_args[2]
        if spline_order == 1: return 0.
        # Form all Gaussian-like integrals necessary
        ll= (numpy.roll(Oparb,-1)[:-1]-self._kick_interpdOpar_poly.c[-1]\
            -self._meandO\
            -self._kick_interpdOpar_poly.c[-2]*self._timpact\
            *(Oparb-numpy.roll(Oparb,-1))[:-1])\
            /numpy.sqrt(2.*self._sortedSigOEig[2])
        ul= (Oparb[:-1]-self._kick_interpdOpar_poly.c[-1]-self._meandO)\
            /numpy.sqrt(2.*self._sortedSigOEig[2])
        gaussxpolyInt=\
            self._densMoments_approx_higherorder_gaussxpolyInts(ll,ul,
                                                               spline_order+2)
        firstTerm= Oparb[:-1]\
            *self._density_par_approx_higherorder(\
            Oparb,lowbindx,_return_array=True,
            gaussxpolyInt=copy.copy(gaussxpolyInt[1:]))
        # Now multiply in the coefficients for each order
        powers= numpy.tile(numpy.arange(spline_order+2)[::-1],
                           (len(ul),1)).T
        gaussxpolyInt*= -0.5*(-numpy.sqrt(2.))**(powers+1)\
            *self._sortedSigOEig[2]**(0.5*(powers-1))
        powers= numpy.tile(numpy.arange(spline_order+1)[::-1][:-2],
                           (len(ul),1)).T
        for jj in range(spline_order+2):
            gaussxpolyInt[-jj-1]*= numpy.sum(\
                self._kick_interpdOpar_poly.c[:-2]
                *self._timpact**powers
                /(1.+self._kick_interpdOpar_poly.c[-2]*self._timpact)
                **(powers+2)
                *special.binom(powers+1,jj)
                *(Oparb[:-1]-self._kick_interpdOpar_poly.c[-1]-self._meandO)
                **(powers-jj+1),axis=0)
        out= numpy.sum(gaussxpolyInt,axis=0)
        out+= firstTerm
        return numpy.sum(out[:lowbindx+1])

    def _determine_deltav_kick(self,impact_angle,impactb,subhalovel,
                               GM,rs,subhalopot,
                               spline_order,hernquist):
        # Store some impact parameters
        self._impactb= impactb
        self._subhalovel= subhalovel
        # Sign of delta angle tells us whether the impact happens to the
        # leading or trailing arm, self._sigMeanSign contains this info;
        # Checked before, but check it again in case impact_angle has changed
        if impact_angle > 0.:
            self._gap_leading= True
        else:
            self._gap_leading= False
        if (self._gap_leading and not self._leading) \
                or (not self._gap_leading and self._leading):
            raise ValueError('Modeling leading (trailing) impact for trailing (leading) arm; this is not allowed because it is nonsensical in this framework')
        self._impact_angle= numpy.fabs(impact_angle)
        # Interpolate the track near the gap in (x,v) at the kick_thetas
        self._interpolate_stream_track_kick()
        self._interpolate_stream_track_kick_aA()
        # Then compute delta v along the track
        if self._general_kick:
            self._kick_deltav=\
                impulse_deltav_general_curvedstream(self._kick_interpolatedObsTrackXY[:,3:],
                                                    self._kick_interpolatedObsTrackXY[:,:3],
                                                    self._impactb,
                                                    self._subhalovel,
                                                    self._kick_ObsTrackXY_closest[:3],
                                                    self._kick_ObsTrackXY_closest[3:],
                                                    subhalopot)
        else:
            if hernquist:
                deltav_func= impulse_deltav_hernquist_curvedstream
            else:
                deltav_func= impulse_deltav_plummer_curvedstream
            self._kick_deltav= \
                deltav_func(self._kick_interpolatedObsTrackXY[:,3:],
                            self._kick_interpolatedObsTrackXY[:,:3],
                            self._impactb,
                            self._subhalovel,
                            self._kick_ObsTrackXY_closest[:3],
                            self._kick_ObsTrackXY_closest[3:],
                            GM,rs)
        return None

    def _determine_deltaOmegaTheta_kick(self,spline_order):
        # Propagate deltav(angle) -> delta (Omega,theta) [angle]
        # Cylindrical coordinates of the perturbed points
        vXp= self._kick_interpolatedObsTrackXY[:,3]+self._kick_deltav[:,0]
        vYp= self._kick_interpolatedObsTrackXY[:,4]+self._kick_deltav[:,1]
        vZp= self._kick_interpolatedObsTrackXY[:,5]+self._kick_deltav[:,2]
        vRp,vTp,vZp=\
            bovy_coords.rect_to_cyl_vec(vXp,vYp,vZp,
                                        self._kick_interpolatedObsTrack[:,0],
                                        self._kick_interpolatedObsTrack[:,5],
                                        self._kick_interpolatedObsTrack[:,3],
                                        cyl=True)
        # We will abuse streamdf functions for doing the (O,a) -> (R,vR)
        # coordinate transformation, to do this, we assign some of the
        # attributes related to the track near the impact to the equivalent
        # attributes related to the track at the present time, carefully
        # removing this again to avoid confusion (as much as possible)
        self._interpolatedObsTrack= self._kick_interpolatedObsTrack
        self._ObsTrack= self._gap_ObsTrack
        self._interpolatedObsTrackXY= self._kick_interpolatedObsTrackXY
        self._ObsTrackXY= self._gap_ObsTrackXY
        self._alljacsTrack= self._gap_alljacsTrack
        self._interpolatedObsTrackAA= self._kick_interpolatedObsTrackAA
        self._ObsTrackAA= self._gap_ObsTrackAA
        self._nTrackChunks= self._nTrackChunksImpact
        Oap= self._approxaA(self._kick_interpolatedObsTrack[:,0],
                            vRp,vTp,
                            self._kick_interpolatedObsTrack[:,3],
                            vZp,
                            self._kick_interpolatedObsTrack[:,5],interp=True,
                            cindx=range(len(self._kick_interpolatedObsTrackAA)))
        # Remove attributes again to avoid confusion later
        delattr(self,'_interpolatedObsTrack')
        delattr(self,'_ObsTrack')
        delattr(self,'_interpolatedObsTrackXY')
        delattr(self,'_ObsTrackXY')
        delattr(self,'_alljacsTrack')
        delattr(self,'_interpolatedObsTrackAA')
        delattr(self,'_ObsTrackAA')
        delattr(self,'_nTrackChunks')
        # Generate (dO,da)[angle_offset] and interpolate (raw here, see below
        # for form that checks range)
        self._kick_dOap= Oap.T-self._kick_interpolatedObsTrackAA
        self._kick_interpdOr_raw=\
            interpolate.InterpolatedUnivariateSpline(\
            self._kick_interpolatedThetasTrack,self._kick_dOap[:,0],
            k=spline_order)
        self._kick_interpdOp_raw=\
            interpolate.InterpolatedUnivariateSpline(\
            self._kick_interpolatedThetasTrack,self._kick_dOap[:,1],
            k=spline_order)
        self._kick_interpdOz_raw=\
            interpolate.InterpolatedUnivariateSpline(\
            self._kick_interpolatedThetasTrack,self._kick_dOap[:,2],
            k=spline_order)
        self._kick_interpdar_raw=\
            interpolate.InterpolatedUnivariateSpline(\
            self._kick_interpolatedThetasTrack,self._kick_dOap[:,3],
            k=spline_order)
        self._kick_interpdap_raw=\
            interpolate.InterpolatedUnivariateSpline(\
            self._kick_interpolatedThetasTrack,self._kick_dOap[:,4],
            k=spline_order)
        self._kick_interpdaz_raw=\
            interpolate.InterpolatedUnivariateSpline(\
            self._kick_interpolatedThetasTrack,self._kick_dOap[:,5],
            k=spline_order)
        # Also interpolate parallel and perpendicular frequencies
        self._kick_dOaparperp=\
            numpy.dot(self._kick_dOap[:,:3],
                      self._sigomatrixEig[1][:,self._sigomatrixEigsortIndx])
        self._kick_dOaparperp[:,2]*= self._sigMeanSign
        self._kick_interpdOpar_raw=\
            interpolate.InterpolatedUnivariateSpline(\
            self._kick_interpolatedThetasTrack,
            numpy.dot(self._kick_dOap[:,:3],self._dsigomeanProgDirection)\
                *self._sigMeanSign,
            k=spline_order) # to get zeros with sproot
        self._kick_interpdOperp0_raw=\
            interpolate.InterpolatedUnivariateSpline(\
            self._kick_interpolatedThetasTrack,self._kick_dOaparperp[:,0],
            k=spline_order)
        self._kick_interpdOperp1_raw=\
            interpolate.InterpolatedUnivariateSpline(\
            self._kick_interpolatedThetasTrack,self._kick_dOaparperp[:,1],
            k=spline_order)
        # Also construct derivative of dOpar
        self._kick_interpdOpar_dapar= self._kick_interpdOpar_raw.derivative(1)
        # Also construct piecewise-polynomial representation of dOpar, 
        # removing intervals at the start and end with zero range
        ppoly= interpolate.PPoly.from_spline(\
            self._kick_interpdOpar_raw._eval_args)
        nzIndx=\
            numpy.nonzero((numpy.roll(ppoly.x,-1)-ppoly.x > 0)\
                              *(numpy.arange(len(ppoly.x)) < len(ppoly.x)//2)\
                              +(ppoly.x-numpy.roll(ppoly.x,1) > 0)\
                              *(numpy.arange(len(ppoly.x)) >= len(ppoly.x)//2))
        self._kick_interpdOpar_poly= interpolate.PPoly(\
            ppoly.c[:,nzIndx[0][:-1]],ppoly.x[nzIndx[0]])
        return None

    # Functions that evaluate the interpolated kicks, but also check the range
    @impact_check_range
    def _kick_interpdOpar(self,da):
        return self._kick_interpdOpar_raw(da)
    @impact_check_range
    def _kick_interpdOperp0(self,da):
        return self._kick_interpdOperp0_raw(da)
    @impact_check_range
    def _kick_interpdOperp1(self,da):
        return self._kick_interpdOperp1_raw(da)
    @impact_check_range
    def _kick_interpdOr(self,da):
        return self._kick_interpdOr_raw(da)
    @impact_check_range
    def _kick_interpdOp(self,da):
        return self._kick_interpdOp_raw(da)
    @impact_check_range
    def _kick_interpdOz(self,da):
        return self._kick_interpdOz_raw(da)
    @impact_check_range
    def _kick_interpdar(self,da):
        return self._kick_interpdar_raw(da)
    @impact_check_range
    def _kick_interpdap(self,da):
        return self._kick_interpdap_raw(da)
    @impact_check_range
    def _kick_interpdaz(self,da):
        return self._kick_interpdaz_raw(da)

    def _interpolate_stream_track_kick(self):
        """Build interpolations of the stream track near the kick"""
        if hasattr(self,'_kick_interpolatedThetasTrack'): #pragma: no cover
            self._store_closest()
            return None #Already did this
        # Setup the trackpoints where the kick will be computed, covering the
        # full length of the stream
        self._kick_interpolatedThetasTrack= \
            numpy.linspace(self._gap_thetasTrack[0],
                           self._gap_thetasTrack[-1],
                           self._nKickPoints)
        TrackX= self._gap_ObsTrack[:,0]*numpy.cos(self._gap_ObsTrack[:,5])
        TrackY= self._gap_ObsTrack[:,0]*numpy.sin(self._gap_ObsTrack[:,5])
        TrackZ= self._gap_ObsTrack[:,3]
        TrackvX, TrackvY, TrackvZ=\
            bovy_coords.cyl_to_rect_vec(self._gap_ObsTrack[:,1],
                                        self._gap_ObsTrack[:,2],
                                        self._gap_ObsTrack[:,4],
                                        self._gap_ObsTrack[:,5])
        #Interpolate
        self._kick_interpTrackX=\
            interpolate.InterpolatedUnivariateSpline(self._gap_thetasTrack,
                                                     TrackX,k=3)
        self._kick_interpTrackY=\
            interpolate.InterpolatedUnivariateSpline(self._gap_thetasTrack,
                                                     TrackY,k=3)
        self._kick_interpTrackZ=\
            interpolate.InterpolatedUnivariateSpline(self._gap_thetasTrack,
                                                     TrackZ,k=3)
        self._kick_interpTrackvX=\
            interpolate.InterpolatedUnivariateSpline(self._gap_thetasTrack,
                                                     TrackvX,k=3)
        self._kick_interpTrackvY=\
            interpolate.InterpolatedUnivariateSpline(self._gap_thetasTrack,
                                                     TrackvY,k=3)
        self._kick_interpTrackvZ=\
            interpolate.InterpolatedUnivariateSpline(self._gap_thetasTrack,
                                                     TrackvZ,k=3)
        #Now store an interpolated version of the stream track
        self._kick_interpolatedObsTrackXY= numpy.empty((len(self._kick_interpolatedThetasTrack),6))
        self._kick_interpolatedObsTrackXY[:,0]=\
            self._kick_interpTrackX(self._kick_interpolatedThetasTrack)
        self._kick_interpolatedObsTrackXY[:,1]=\
            self._kick_interpTrackY(self._kick_interpolatedThetasTrack)
        self._kick_interpolatedObsTrackXY[:,2]=\
            self._kick_interpTrackZ(self._kick_interpolatedThetasTrack)
        self._kick_interpolatedObsTrackXY[:,3]=\
            self._kick_interpTrackvX(self._kick_interpolatedThetasTrack)
        self._kick_interpolatedObsTrackXY[:,4]=\
            self._kick_interpTrackvY(self._kick_interpolatedThetasTrack)
        self._kick_interpolatedObsTrackXY[:,5]=\
            self._kick_interpTrackvZ(self._kick_interpolatedThetasTrack)
        #Also in cylindrical coordinates
        self._kick_interpolatedObsTrack= \
            numpy.empty((len(self._kick_interpolatedThetasTrack),6))
        tR,tphi,tZ= bovy_coords.rect_to_cyl(self._kick_interpolatedObsTrackXY[:,0],
                                            self._kick_interpolatedObsTrackXY[:,1],
                                            self._kick_interpolatedObsTrackXY[:,2])
        tvR,tvT,tvZ=\
            bovy_coords.rect_to_cyl_vec(self._kick_interpolatedObsTrackXY[:,3],
                                        self._kick_interpolatedObsTrackXY[:,4],
                                        self._kick_interpolatedObsTrackXY[:,5],
                                        tR,tphi,tZ,cyl=True)
        self._kick_interpolatedObsTrack[:,0]= tR
        self._kick_interpolatedObsTrack[:,1]= tvR
        self._kick_interpolatedObsTrack[:,2]= tvT
        self._kick_interpolatedObsTrack[:,3]= tZ
        self._kick_interpolatedObsTrack[:,4]= tvZ
        self._kick_interpolatedObsTrack[:,5]= tphi
        self._store_closest()
        return None

    def _store_closest(self):
        # Also store (x,v) for the point of closest approach
        self._kick_ObsTrackXY_closest= numpy.array([\
                self._kick_interpTrackX(self._impact_angle),
                self._kick_interpTrackY(self._impact_angle),
                self._kick_interpTrackZ(self._impact_angle),
                self._kick_interpTrackvX(self._impact_angle),
                self._kick_interpTrackvY(self._impact_angle),
                self._kick_interpTrackvZ(self._impact_angle)])
        return None

    def _interpolate_stream_track_kick_aA(self):
        """Build interpolations of the stream track near the impact in action-angle coordinates"""
        if hasattr(self,'_kick_interpolatedObsTrackAA'): #pragma: no cover
            return None #Already did this
        #Calculate 1D meanOmega on a fine grid in angle and interpolate
        dmOs= numpy.array([\
                super(streamgapdf,self).meanOmega(da,oned=True,
                                                  tdisrupt=self._tdisrupt
                                                  -self._timpact,
                                                  use_physical=False)
                for da in self._kick_interpolatedThetasTrack])
        self._kick_interpTrackAAdmeanOmegaOneD=\
            interpolate.InterpolatedUnivariateSpline(\
            self._kick_interpolatedThetasTrack,dmOs,k=3)
        #Build the interpolated AA
        self._kick_interpolatedObsTrackAA=\
            numpy.empty((len(self._kick_interpolatedThetasTrack),6))
        for ii in range(len(self._kick_interpolatedThetasTrack)):
            self._kick_interpolatedObsTrackAA[ii,:3]=\
                self._progenitor_Omega+dmOs[ii]*self._dsigomeanProgDirection\
                *self._gap_sigMeanSign
            self._kick_interpolatedObsTrackAA[ii,3:]=\
                self._progenitor_angle+self._kick_interpolatedThetasTrack[ii]\
                *self._dsigomeanProgDirection*self._gap_sigMeanSign\
                -self._timpact*self._progenitor_Omega
            self._kick_interpolatedObsTrackAA[ii,3:]=\
                numpy.mod(self._kick_interpolatedObsTrackAA[ii,3:],2.*numpy.pi)
        return None

    def _determine_deltaAngleTrackImpact(self,deltaAngleTrackImpact,timpact):
        self._timpact= timpact
        deltaAngleTrackLim = (self._sigMeanOffset+4.) * numpy.sqrt(
            self._sortedSigOEig[2]) * (self._tdisrupt-self._timpact)
        if deltaAngleTrackImpact is None:
            deltaAngleTrackImpact= deltaAngleTrackLim
        else:
            if deltaAngleTrackImpact > deltaAngleTrackLim:
                warnings.warn("WARNING: deltaAngleTrackImpact angle range large compared to plausible value", galpyWarning)
        self._deltaAngleTrackImpact= deltaAngleTrackImpact
        return None

    def _determine_impact_coordtransform(self,deltaAngleTrackImpact,
                                         nTrackChunksImpact,
                                         timpact,impact_angle):
        """Function that sets up the transformation between (x,v) and (O,theta)"""
        # Integrate the progenitor backward to the time of impact
        self._gap_progenitor_setup()
        # Sign of delta angle tells us whether the impact happens to the
        # leading or trailing arm, self._sigMeanSign contains this info
        if impact_angle > 0.:
            self._gap_leading= True
        else:
            self._gap_leading= False
        if (self._gap_leading and not self._leading) \
                or (not self._gap_leading and self._leading):
            raise ValueError('Modeling leading (trailing) impact for trailing (leading) arm; this is not allowed because it is nonsensical in this framework')
        self._gap_sigMeanSign= 1.
        if (self._gap_leading and self._progenitor_Omega_along_dOmega/self._sigMeanSign < 0.) \
                or (not self._gap_leading and self._progenitor_Omega_along_dOmega/self._sigMeanSign > 0.):
            self._gap_sigMeanSign= -1.
        # Determine how much orbital time is necessary for the progenitor's orbit at the time of impact to cover the part of the stream near the impact; we cover the whole leading (or trailing) part of the stream
        if nTrackChunksImpact is None:
            #default is floor(self._deltaAngleTrackImpact/0.15)+1
            self._nTrackChunksImpact= int(numpy.floor(self._deltaAngleTrackImpact/0.15))+1
        else:
            self._nTrackChunksImpact= nTrackChunksImpact
        if self._nTrackChunksImpact < 4: self._nTrackChunksImpact= 4
        dt= self._deltaAngleTrackImpact\
            /self._progenitor_Omega_along_dOmega\
            /self._sigMeanSign*self._gap_sigMeanSign
        self._gap_trackts= numpy.linspace(0.,2*dt,2*self._nTrackChunksImpact-1) #to be sure that we cover it
        #Instantiate an auxiliaryTrack, which is an Orbit instance at the mean frequency of the stream, and zero angle separation wrt the progenitor; prog_stream_offset is the offset between this track and the progenitor at zero angle (same as in streamdf, but just done at the time of impact rather than the current time)
        prog_stream_offset=\
            _determine_stream_track_single(self._aA,
                                           self._gap_progenitor,
                                           self._timpact, # around the t of imp
                                           self._progenitor_angle-self._timpact*self._progenitor_Omega,
                                           self._gap_sigMeanSign,
                                           self._dsigomeanProgDirection,
                                           lambda da: super(streamgapdf,self).meanOmega(da,offset_sign=self._gap_sigMeanSign,tdisrupt=self._tdisrupt-self._timpact,use_physical=False),
                                           0.) #angle = 0
        auxiliaryTrack= Orbit(prog_stream_offset[3])
        if dt < 0.:
            self._gap_trackts= numpy.linspace(0.,-2.*dt,2.*self._nTrackChunksImpact-1)
            #Flip velocities before integrating
            auxiliaryTrack= auxiliaryTrack.flip()
        auxiliaryTrack.integrate(self._gap_trackts,self._pot)
        if dt < 0.:
            #Flip velocities again
            auxiliaryTrack.orbit[...,1]= -auxiliaryTrack.orbit[...,1]
            auxiliaryTrack.orbit[...,2]= -auxiliaryTrack.orbit[...,2]
            auxiliaryTrack.orbit[...,4]= -auxiliaryTrack.orbit[...,4]
        #Calculate the actions, frequencies, and angle for this auxiliary orbit
        acfs= self._aA.actionsFreqs(auxiliaryTrack(0.),maxn=3,
                                    use_physical=False)
        auxiliary_Omega= numpy.array([acfs[3],acfs[4],acfs[5]]).reshape(3\
)
        auxiliary_Omega_along_dOmega= \
            numpy.dot(auxiliary_Omega,self._dsigomeanProgDirection)
        # compute the transformation using _determine_stream_track_single
        allAcfsTrack= numpy.empty((self._nTrackChunksImpact,9))
        alljacsTrack= numpy.empty((self._nTrackChunksImpact,6,6))
        allinvjacsTrack= numpy.empty((self._nTrackChunksImpact,6,6))
        thetasTrack= numpy.linspace(0.,self._deltaAngleTrackImpact,
                                    self._nTrackChunksImpact)
        ObsTrack= numpy.empty((self._nTrackChunksImpact,6))
        ObsTrackAA= numpy.empty((self._nTrackChunksImpact,6))
        detdOdJps= numpy.empty((self._nTrackChunksImpact))
        if self._multi is None:
            for ii in range(self._nTrackChunksImpact):
                multiOut= _determine_stream_track_single(self._aA,
                                           auxiliaryTrack,
                                           self._gap_trackts[ii]*numpy.fabs(self._progenitor_Omega_along_dOmega/auxiliary_Omega_along_dOmega), #this factor accounts for the difference in frequency between the progenitor and the auxiliary track, no timpact bc gap_tracks is relative to timpact
                                           self._progenitor_angle-self._timpact*self._progenitor_Omega,
                                           self._gap_sigMeanSign,
                                           self._dsigomeanProgDirection,
                                           lambda da: super(streamgapdf,self).meanOmega(da,offset_sign=self._gap_sigMeanSign,tdisrupt=self._tdisrupt-self._timpact,use_physical=False),
                                           thetasTrack[ii])
                allAcfsTrack[ii,:]= multiOut[0]
                alljacsTrack[ii,:,:]= multiOut[1]
                allinvjacsTrack[ii,:,:]= multiOut[2]
                ObsTrack[ii,:]= multiOut[3]
                ObsTrackAA[ii,:]= multiOut[4]
                detdOdJps[ii]= multiOut[5]
        else:
            multiOut= multi.parallel_map(\
                (lambda x: _determine_stream_track_single(self._aA,
                                           auxiliaryTrack,
                                           self._gap_trackts[x]*numpy.fabs(self._progenitor_Omega_along_dOmega/auxiliary_Omega_along_dOmega), #this factor accounts for the difference in frequency between the progenitor and the auxiliary track, no timpact bc gap_tracks is relative to timpact
                                           self._progenitor_angle-self._timpact*self._progenitor_Omega,
                                           self._gap_sigMeanSign,
                                           self._dsigomeanProgDirection,
                                           lambda da: super(streamgapdf,self).meanOmega(da,offset_sign=self._gap_sigMeanSign,tdisrupt=self._tdisrupt-self._timpact,use_physical=False),
                                           thetasTrack[x])),
                range(self._nTrackChunksImpact),
                numcores=numpy.amin([self._nTrackChunksImpact,
                                     multiprocessing.cpu_count(),
                                     self._multi]))
            for ii in range(self._nTrackChunksImpact):
                allAcfsTrack[ii,:]= multiOut[ii][0]
                alljacsTrack[ii,:,:]= multiOut[ii][1]
                allinvjacsTrack[ii,:,:]= multiOut[ii][2]
                ObsTrack[ii,:]= multiOut[ii][3]
                ObsTrackAA[ii,:]= multiOut[ii][4]
                detdOdJps[ii]= multiOut[ii][5]
        #Repeat the track calculation using the previous track, to get closer to it
        for nn in range(self.nTrackIterations):
            if self._multi is None:
                for ii in range(self._nTrackChunksImpact):
                    multiOut= _determine_stream_track_single(self._aA,
                                                             Orbit(ObsTrack[ii,:]),
                                                             0.,
                                                             self._progenitor_angle-self._timpact*self._progenitor_Omega,
                                                             self._gap_sigMeanSign,
                                                             self._dsigomeanProgDirection,
                                                             lambda da: super(streamgapdf,self).meanOmega(da,offset_sign=self._gap_sigMeanSign,tdisrupt=self._tdisrupt-self._timpact,use_physical=False),
                                                             thetasTrack[ii])
                    allAcfsTrack[ii,:]= multiOut[0]
                    alljacsTrack[ii,:,:]= multiOut[1]
                    allinvjacsTrack[ii,:,:]= multiOut[2]
                    ObsTrack[ii,:]= multiOut[3]
                    ObsTrackAA[ii,:]= multiOut[4]
                    detdOdJps[ii]= multiOut[5]
            else:
                multiOut= multi.parallel_map(\
                    (lambda x: _determine_stream_track_single(self._aA,Orbit(ObsTrack[x,:]),0.,
                                                              self._progenitor_angle-self._timpact*self._progenitor_Omega,
                                                              self._gap_sigMeanSign,
                                                              self._dsigomeanProgDirection,
                                                              lambda da: super(streamgapdf,self).meanOmega(da,offset_sign=self._gap_sigMeanSign,tdisrupt=self._tdisrupt-self._timpact,use_physical=False),
                                                              thetasTrack[x])),
                    range(self._nTrackChunksImpact),
                    numcores=numpy.amin([self._nTrackChunksImpact,
                                         multiprocessing.cpu_count(),
                                         self._multi]))
                for ii in range(self._nTrackChunksImpact):
                    allAcfsTrack[ii,:]= multiOut[ii][0]
                    alljacsTrack[ii,:,:]= multiOut[ii][1]
                    allinvjacsTrack[ii,:,:]= multiOut[ii][2]
                    ObsTrack[ii,:]= multiOut[ii][3]
                    ObsTrackAA[ii,:]= multiOut[ii][4]
                    detdOdJps[ii]= multiOut[ii][5]
        #Store the track
        self._gap_thetasTrack= thetasTrack
        self._gap_ObsTrack= ObsTrack
        self._gap_ObsTrackAA= ObsTrackAA
        self._gap_allAcfsTrack= allAcfsTrack
        self._gap_alljacsTrack= alljacsTrack
        self._gap_allinvjacsTrack= allinvjacsTrack
        self._gap_detdOdJps= detdOdJps
        self._gap_meandetdOdJp= numpy.mean(self._gap_detdOdJps)
        self._gap_logmeandetdOdJp= numpy.log(self._gap_meandetdOdJp)
        #Also calculate _ObsTrackXY in XYZ,vXYZ coordinates
        self._gap_ObsTrackXY= numpy.empty_like(self._gap_ObsTrack)
        TrackX= self._gap_ObsTrack[:,0]*numpy.cos(self._gap_ObsTrack[:,5])
        TrackY= self._gap_ObsTrack[:,0]*numpy.sin(self._gap_ObsTrack[:,5])
        TrackZ= self._gap_ObsTrack[:,3]
        TrackvX, TrackvY, TrackvZ=\
            bovy_coords.cyl_to_rect_vec(self._gap_ObsTrack[:,1],
                                        self._gap_ObsTrack[:,2],
                                        self._gap_ObsTrack[:,4],
                                        self._gap_ObsTrack[:,5])
        self._gap_ObsTrackXY[:,0]= TrackX
        self._gap_ObsTrackXY[:,1]= TrackY
        self._gap_ObsTrackXY[:,2]= TrackZ
        self._gap_ObsTrackXY[:,3]= TrackvX
        self._gap_ObsTrackXY[:,4]= TrackvY
        self._gap_ObsTrackXY[:,5]= TrackvZ
        return None

    def _gap_progenitor_setup(self):
        """Setup an Orbit instance that's the progenitor integrated backwards"""
        self._gap_progenitor= self._progenitor().flip() # new orbit, flip velocities
        # Make sure we do not use physical coordinates
        self._gap_progenitor.turn_physical_off()
        # Now integrate backward in time until tdisrupt
        ts= numpy.linspace(0.,self._tdisrupt,1001)
        self._gap_progenitor.integrate(ts,self._pot)
        # Flip its velocities, should really write a function for this
        self._gap_progenitor.orbit[...,1]= -self._gap_progenitor.orbit[...,1]
        self._gap_progenitor.orbit[...,2]= -self._gap_progenitor.orbit[...,2]
        self._gap_progenitor.orbit[...,4]= -self._gap_progenitor.orbit[...,4]
        return None

################################SAMPLE THE DF##################################
    def _sample_aAt(self,n):
        """Sampling frequencies, angles, and times part of sampling, for stream with gap"""
        # Use streamdf's _sample_aAt to generate unperturbed frequencies,
        # angles
        Om,angle,dt= super(streamgapdf,self)._sample_aAt(n)
        # Now rewind angles by timpact, apply the kicks, and run forward again
        dangle_at_impact= angle-numpy.tile(self._progenitor_angle.T,(n,1)).T\
            -(Om-numpy.tile(self._progenitor_Omega.T,(n,1)).T)*self._timpact
        dangle_par_at_impact= numpy.dot(dangle_at_impact.T,
                                        self._dsigomeanProgDirection)\
                                        *self._gap_sigMeanSign
        # Calculate and apply kicks (points not yet released have zero kick)
        dOr= self._kick_interpdOr(dangle_par_at_impact)
        dOp= self._kick_interpdOp(dangle_par_at_impact)
        dOz= self._kick_interpdOz(dangle_par_at_impact)
        Om[0,:]+= dOr
        Om[1,:]+= dOp
        Om[2,:]+= dOz
        angle[0,:]+=\
            self._kick_interpdar(dangle_par_at_impact)+dOr*self._timpact
        angle[1,:]+=\
            self._kick_interpdap(dangle_par_at_impact)+dOp*self._timpact
        angle[2,:]+=\
            self._kick_interpdaz(dangle_par_at_impact)+dOz*self._timpact
        return (Om,angle,dt)

def impulse_deltav_plummer(v,y,b,w,GM,rs):
    """
    NAME:

       impulse_deltav_plummer

    PURPOSE:

       calculate the delta velocity to due an encounter with a Plummer sphere in the impulse approximation; allows for arbitrary velocity vectors, but y is input as the position along the stream

    INPUT:

       v - velocity of the stream (nstar,3)

       y - position along the stream (nstar)

       b - impact parameter

       w - velocity of the Plummer sphere (3)

       GM - mass of the Plummer sphere (in natural units)

       rs - size of the Plummer sphere

    OUTPUT:

       deltav (nstar,3)

    HISTORY:

       2015-04-30 - Written based on Erkal's expressions - Bovy (IAS)

    """
    if len(v.shape) == 1:
        v= numpy.reshape(v,(1,3))
        y= numpy.reshape(y,(1,1))
    nv= v.shape[0]
    # Build the rotation matrices and their inverse
    rot= _rotation_vy(v)
    rotinv= _rotation_vy(v,inv=True)
    # Rotate the Plummer sphere's velocity to the stream frames
    tilew= numpy.sum(rot*numpy.tile(w,(nv,3,1)),axis=-1)
    # Use Denis' expressions
    wperp= numpy.sqrt(tilew[:,0]**2.+tilew[:,2]**2.)
    wpar= numpy.sqrt(numpy.sum(v**2.,axis=1))-tilew[:,1]
    wmag2= wpar**2.+wperp**2.
    wmag= numpy.sqrt(wmag2)
    out= numpy.empty_like(v)
    denom= wmag*((b**2.+rs**2.)*wmag2+wperp**2.*y**2.)
    out[:,0]= (b*wmag2*tilew[:,2]/wperp-y*wpar*tilew[:,0])/denom
    out[:,1]= -wperp**2.*y/denom
    out[:,2]= -(b*wmag2*tilew[:,0]/wperp+y*wpar*tilew[:,2])/denom
    # deal w/ perpendicular impacts
    wperp0Indx= numpy.fabs(wperp) < 10.**-10.
    out[wperp0Indx,0]= (b*wmag2[wperp0Indx]-y[wperp0Indx]*wpar[wperp0Indx]*tilew[wperp0Indx,0])/denom[wperp0Indx]
    out[wperp0Indx,2]=-(b*wmag2[wperp0Indx]+y[wperp0Indx]*wpar[wperp0Indx]*tilew[wperp0Indx,2])/denom[wperp0Indx]
    # Rotate back to the original frame
    return 2.0*GM*numpy.sum(\
        rotinv*numpy.swapaxes(numpy.tile(out.T,(3,1,1)).T,1,2),axis=-1)

def impulse_deltav_plummer_curvedstream(v,x,b,w,x0,v0,GM,rs):
    """
    NAME:

       impulse_deltav_plummer_curvedstream

    PURPOSE:

       calculate the delta velocity to due an encounter with a Plummer sphere in the impulse approximation; allows for arbitrary velocity vectors, and arbitrary position along the stream

    INPUT:

       v - velocity of the stream (nstar,3)

       x - position along the stream (nstar,3)

       b - impact parameter

       w - velocity of the Plummer sphere (3)

       x0 - point of closest approach

       v0 - velocity of point of closest approach

       GM - mass of the Plummer sphere (in natural units)

       rs - size of the Plummer sphere

    OUTPUT:

       deltav (nstar,3)

    HISTORY:

       2015-05-04 - Written based on above - SANDERS

    """
    if len(v.shape) == 1: v= numpy.reshape(v,(1,3))
    if len(x.shape) == 1: x= numpy.reshape(x,(1,3))
    b0 = numpy.cross(w,v0)
    b0 *= b/numpy.sqrt(numpy.sum(b0**2))
    b_ = b0+x-x0
    w = w-v
    wmag = numpy.sqrt(numpy.sum(w**2,axis=1))
    bdotw=numpy.sum(b_*w,axis=1)/wmag
    denom= wmag*(numpy.sum(b_**2,axis=1)+rs**2-bdotw**2)
    denom = 1./denom
    return -2.0*GM*((b_.T-bdotw*w.T/wmag)*denom).T

def HernquistX(s):
    """
      Computes X function from equations (33) & (34) of Hernquist (1990)
    """
    if(s<0.):
        raise ValueError("s must be positive in Hernquist X function")
    elif(s<1.):
        return numpy.log((1+numpy.sqrt(1-s*s))/s)/numpy.sqrt(1-s*s)
    elif(s==1.):
        return 1.
    else:
        return numpy.arccos(1./s)/numpy.sqrt(s*s-1)

def impulse_deltav_hernquist(v,y,b,w,GM,rs):
    """
    NAME:

       impulse_deltav_hernquist

    PURPOSE:

       calculate the delta velocity to due an encounter with a Hernquist sphere in the impulse approximation; allows for arbitrary velocity vectors, but y is input as the position along the stream

    INPUT:

       v - velocity of the stream (nstar,3)

       y - position along the stream (nstar)

       b - impact parameter

       w - velocity of the Hernquist sphere (3)

       GM - mass of the Hernquist sphere (in natural units)

       rs - size of the Hernquist sphere

    OUTPUT:

       deltav (nstar,3)

    HISTORY:

       2015-08-13 SANDERS, using Wyn Evans calculation

    """
    if len(v.shape) == 1: v= numpy.reshape(v,(1,3))
    nv= v.shape[0]
    # Build the rotation matrices and their inverse
    rot= _rotation_vy(v)
    rotinv= _rotation_vy(v,inv=True)
    # Rotate the Plummer sphere's velocity to the stream frames
    tilew= numpy.sum(rot*numpy.tile(w,(nv,3,1)),axis=-1)
    wperp= numpy.sqrt(tilew[:,0]**2.+tilew[:,2]**2.)
    wpar= numpy.sqrt(numpy.sum(v**2.,axis=1))-tilew[:,1]
    wmag2= wpar**2.+wperp**2.
    wmag= numpy.sqrt(wmag2)
    B = numpy.sqrt(b**2.+wperp**2.*y**2./wmag2)
    denom= wmag*(B**2-rs**2)
    denom=1./denom
    s = numpy.sqrt(2.*B/(rs+B))
    HernquistXv=numpy.vectorize(HernquistX)
    Xfac = 1.-2.*rs/(rs+B)*HernquistXv(s)
    out= numpy.empty_like(v)
    out[:,0]= (b*tilew[:,2]/wperp-y*wpar*tilew[:,0]/wmag2)*denom*Xfac
    out[:,1]= -wperp**2.*y*denom*Xfac/wmag2
    out[:,2]= -(b*tilew[:,0]/wperp+y*wpar*tilew[:,2]/wmag2)*denom*Xfac
    # deal w/ perpendicular impacts
    wperp0Indx= numpy.fabs(wperp) < 10.**-10.
    out[wperp0Indx,0]= (b-y[wperp0Indx]*wpar[wperp0Indx]*tilew[wperp0Indx,0]/wmag2[wperp0Indx])*denom[wperp0Indx]*Xfac[wperp0Indx]
    out[wperp0Indx,2]=-(b+y[wperp0Indx]*wpar[wperp0Indx]*tilew[wperp0Indx,2]/wmag2[wperp0Indx])*denom[wperp0Indx]*Xfac[wperp0Indx]
    # Rotate back to the original frame
    return 2.0*GM*numpy.sum(\
        rotinv*numpy.swapaxes(numpy.tile(out.T,(3,1,1)).T,1,2),axis=-1)

def impulse_deltav_hernquist_curvedstream(v,x,b,w,x0,v0,GM,rs):
    """
    NAME:

       impulse_deltav_plummer_hernquist

    PURPOSE:

       calculate the delta velocity to due an encounter with a Hernquist sphere in the impulse approximation; allows for arbitrary velocity vectors, and arbitrary position along the stream

    INPUT:

       v - velocity of the stream (nstar,3)

       x - position along the stream (nstar,3)

       b - impact parameter

       w - velocity of the Hernquist sphere (3)

       x0 - point of closest approach

       v0 - velocity of point of closest approach

       GM - mass of the Hernquist sphere (in natural units)

       rs - size of the Hernquist sphere

    OUTPUT:

       deltav (nstar,3)

    HISTORY:

       2015-08-13 - SANDERS, using Wyn Evans calculation

    """
    if len(v.shape) == 1: v= numpy.reshape(v,(1,3))
    if len(x.shape) == 1: x= numpy.reshape(x,(1,3))
    b0 = numpy.cross(w,v0)
    b0 *= b/numpy.sqrt(numpy.sum(b0**2))
    b_ = b0+x-x0
    w = w-v
    wmag = numpy.sqrt(numpy.sum(w**2,axis=1))
    bdotw=numpy.sum(b_*w,axis=1)/wmag
    B = numpy.sqrt(numpy.sum(b_**2,axis=1)-bdotw**2)
    denom= wmag*(B**2-rs**2)
    denom=1./denom
    s = numpy.sqrt(2.*B/(rs+B))
    HernquistXv=numpy.vectorize(HernquistX)
    Xfac = 1.-2.*rs/(rs+B)*HernquistXv(s)
    return -2.0*GM*((b_.T-bdotw*w.T/wmag)*Xfac*denom).T

def _a_integrand(T,y,b,w,pot,compt):
    t = T/(1-T*T)
    X = b+w*t+y*numpy.array([0,1,0])
    r = numpy.sqrt(numpy.sum(X**2))
    return (1+T*T)/(1-T*T)**2*evaluateRforces(pot,r,0.)*X[compt]/r

def _deltav_integrate(y,b,w,pot):
    return numpy.array([integrate.quad(_a_integrand,-1.,1.,args=(y,b,w,pot,i))[0] for i in range(3)])

def impulse_deltav_general(v,y,b,w,pot):
    """
    NAME:

       impulse_deltav_general


    PURPOSE:

       calculate the delta velocity to due an encounter with a general spherical potential in the impulse approximation; allows for arbitrary velocity vectors, but y is input as the position along the stream

    INPUT:

       v - velocity of the stream (nstar,3)

       y - position along the stream (nstar)

       b - impact parameter

       w - velocity of the subhalo (3)

       pot - Potential object or list thereof (should be spherical)

    OUTPUT:

       deltav (nstar,3)

    HISTORY:

       2015-05-04 - SANDERS

       2015-06-15 - Tweak to use galpy' potential objects - Bovy (IAS)

    """
    pot= flatten_potential(pot)
    if len(v.shape) == 1: v= numpy.reshape(v,(1,3))
    nv= v.shape[0]
    # Build the rotation matrices and their inverse
    rot= _rotation_vy(v)
    rotinv= _rotation_vy(v,inv=True)
    # Rotate the subhalo's velocity to the stream frames
    tilew= numpy.sum(rot*numpy.tile(w,(nv,3,1)),axis=-1)
    tilew[:,1]-=numpy.sqrt(numpy.sum(v**2.,axis=1))
    wmag = numpy.sqrt(tilew[:,0]**2+tilew[:,2]**2)
    b0 = b*numpy.array([-tilew[:,2]/wmag,numpy.zeros(nv),tilew[:,0]/wmag]).T
    return numpy.array(list(map(lambda i:numpy.sum(i[3]
                       *_deltav_integrate(i[0],i[1],i[2],pot).T,axis=-1)
                        ,zip(y,b0,tilew,rotinv))))

def impulse_deltav_general_curvedstream(v,x,b,w,x0,v0,pot):
    """
    NAME:

       impulse_deltav_general_curvedstream

    PURPOSE:

       calculate the delta velocity to due an encounter with a general spherical potential in the impulse approximation; allows for arbitrary velocity vectors and arbitrary shaped streams

    INPUT:

       v - velocity of the stream (nstar,3)

       x - position along the stream (nstar,3)

       b - impact parameter

       w - velocity of the subhalo (3)

       x0 - position of closest approach (3)

       v0 - velocity of stream at closest approach (3)

       pot - Potential object or list thereof (should be spherical)

    OUTPUT:

       deltav (nstar,3)

    HISTORY:

       2015-05-04 - SANDERS

       2015-06-15 - Tweak to use galpy' potential objects - Bovy (IAS)

    """
    pot= flatten_potential(pot)
    if len(v.shape) == 1: v= numpy.reshape(v,(1,3))
    if len(x.shape) == 1: x= numpy.reshape(x,(1,3))
    b0 = numpy.cross(w,v0)
    b0 *= b/numpy.sqrt(numpy.sum(b0**2))
    b_ = b0+x-x0
    return numpy.array(list(map(lambda i:_deltav_integrate(0.,i[1],i[0],pot)
                        ,zip(w-v,b_))))

def impulse_deltav_general_orbitintegration(v,x,b,w,x0,v0,pot,tmax,galpot,
                                            tmaxfac=10.,nsamp=1000,
                                            integrate_method='symplec4_c'):
    """
    NAME:

       impulse_deltav_general_orbitintegration

    PURPOSE:

       calculate the delta velocity to due an encounter with a general spherical potential NOT in the impulse approximation by integrating each particle in the underlying galactic potential; allows for arbitrary velocity vectors and arbitrary shaped streams.

    INPUT:

       v - velocity of the stream (nstar,3)

       x - position along the stream (nstar,3)

       b - impact parameter

       w - velocity of the subhalo (3)

       x0 - position of closest approach (3)

       v0 - velocity of stream at closest approach (3)

       pot - Potential object or list thereof (should be spherical)

       tmax - maximum integration time

       galpot - galpy Potential object or list thereof

       nsamp(1000) - number of forward integration points

       integrate_method= ('symplec4_c') orbit integrator to use (see Orbit.integrate)

    OUTPUT:

       deltav (nstar,3)

    HISTORY:

       2015-08-17 - SANDERS

    """
    galpot= flatten_potential(galpot)
    if len(v.shape) == 1: v= numpy.reshape(v,(1,3))
    if len(x.shape) == 1: x= numpy.reshape(x,(1,3))
    nstar,ndim=numpy.shape(v)
    b0 = numpy.cross(w,v0)
    b0 *= b/numpy.sqrt(numpy.sum(b0**2))
    times = numpy.linspace(0.,tmax,nsamp)
    xres = numpy.zeros(shape=(len(x),nsamp*2-1,3))
    R, phi, z= bovy_coords.rect_to_cyl(x[:,0],x[:,1],x[:,2])
    vR, vp, vz= bovy_coords.rect_to_cyl_vec(v[:,0],v[:,1],v[:,2],
                                            R,phi,z,cyl=True)
    for i in range(nstar):
        o = Orbit([R[i],vR[i],vp[i],z[i],vz[i],phi[i]])
        o.integrate(times,galpot,method=integrate_method)
        xres[i,nsamp:,0]=o.x(times)[1:]
        xres[i,nsamp:,1]=o.y(times)[1:]
        xres[i,nsamp:,2]=o.z(times)[1:]
        oreverse = o.flip()
        oreverse.integrate(times,galpot,method=integrate_method)
        xres[i,:nsamp,0]=oreverse.x(times)[::-1]
        xres[i,:nsamp,1]=oreverse.y(times)[::-1]
        xres[i,:nsamp,2]=oreverse.z(times)[::-1]
    times = numpy.concatenate((-times[::-1],times[1:]))
    nsamp = len(times)
    X = b0+xres-x0-numpy.outer(times,w)
    r = numpy.sqrt(numpy.sum(X**2,axis=-1))
    acc = (numpy.reshape(evaluateRforces(pot,r.flatten(),0.),(nstar,nsamp))/r)[:,:,numpy.newaxis]*X
    return integrate.simps(acc,x=times,axis=1)

def impulse_deltav_general_fullplummerintegration(v,x,b,w,x0,v0,galpot,GM,rs,
                                                  tmaxfac=10.,N=1000,
                                                integrate_method='symplec4_c'):
    """
    NAME:

       impulse_deltav_general_fullplummerintegration

    PURPOSE:

       calculate the delta velocity to due an encounter with a moving Plummer sphere and galactic potential relative to just in galactic potential

    INPUT:

       v - velocity of the stream (nstar,3)

       x - position along the stream (nstar,3)

       b - impact parameter

       w - velocity of the subhalo (3)

       x0 - position of closest approach (3)

       v0 - velocity of stream at closest approach (3)

       galpot - Galaxy Potential object

       GM - mass of Plummer

       rs - scale of Plummer

       tmaxfac(10) - multiple of rs/fabs(w - v0) to use for time integration interval

       N(1000) - number of forward integration points

       integrate_method('symplec4_c') - orbit integrator to use (see Orbit.integrate)

    OUTPUT:

       deltav (nstar,3)

    HISTORY:

       2015-08-18 - SANDERS

    """
    galpot= flatten_potential(galpot)
    if len(v.shape) == 1: v= numpy.reshape(v,(1,3))
    if len(x.shape) == 1: x= numpy.reshape(x,(1,3))

    nstar,ndim=numpy.shape(v)
    b0 = numpy.cross(w,v0)
    b0 *= b/numpy.sqrt(numpy.sum(b0**2))
    X = x0-b0

    # Setup Plummer orbit
    R, phi, z= bovy_coords.rect_to_cyl(X[0],X[1],X[2])
    vR, vp, vz= bovy_coords.rect_to_cyl_vec(w[0],w[1],w[2],R,phi,z,cyl=True)
    tmax = tmaxfac*rs/numpy.sqrt(numpy.sum((w-v0)**2))
    times = numpy.linspace(0.,tmax,N)
    dtimes = numpy.linspace(-tmax,tmax,2*N)
    o = Orbit(vxvv=[R,-vR,-vp,z,-vz,phi])
    o.integrate(times,galpot,method=integrate_method)
    oplum = o(times[-1]).flip()
    oplum.integrate(dtimes,galpot,method=integrate_method)
    plumpot = MovingObjectPotential(orbit=oplum,
                                    pot=PlummerPotential(amp=GM,b=rs))

    # Now integrate each particle backwards in galaxy potential, forwards in combined potential and backwards again in galaxy and take diff

    deltav = numpy.zeros((nstar,3))
    R, phi, z= bovy_coords.rect_to_cyl(x[:,0],x[:,1],x[:,2])
    vR, vp, vz= bovy_coords.rect_to_cyl_vec(v[:,0],v[:,1],v[:,2],
                                            R,phi,z,cyl=True)
    for i in range(nstar):
      ostar= Orbit(vxvv=[R[i],-vR[i],-vp[i],z[i],-vz[i],phi[i]])
      ostar.integrate(times,galpot,method=integrate_method)
      oboth = ostar(times[-1]).flip()
      oboth.integrate(dtimes,[galpot,plumpot],method=integrate_method)
      ogalpot = oboth(times[-1]).flip()
      ogalpot.integrate(times,galpot,method=integrate_method)
      deltav[i][0] = -ogalpot.vx(times[-1]) - v[i][0]
      deltav[i][1] = -ogalpot.vy(times[-1]) - v[i][1]
      deltav[i][2] = -ogalpot.vz(times[-1]) - v[i][2]
    return deltav

def _astream_integrand_x(t,y,v,b,w,b2,w2,wperp,wperp2,wpar,GSigma,rs2):
    return GSigma(t)*(b*w2*w[2]/wperp-(y-v*t)*wpar*w[0])\
        /((b2+rs2)*w2+wperp2*(y-v*t)**2.)
def _astream_integrand_y(t,y,v,b2,w2,wperp2,GSigma,rs2):
    return GSigma(t)*(y-v*t)/((b2+rs2)*w2+wperp2*(y-v*t)**2.)
def _astream_integrand_z(t,y,v,b,w,b2,w2,wperp,wperp2,wpar,GSigma,rs2):
    return -GSigma(t)*(b*w2*w[0]/wperp+(y-v*t)*wpar*w[2])\
        /((b2+rs2)*w2+wperp2*(y-v*t)**2.)

def impulse_deltav_plummerstream(v,y,b,w,GSigma,rs,tmin=None,tmax=None):
    """
    NAME:

       impulse_deltav_plummerstream

    PURPOSE:

       calculate the delta velocity to due an encounter with a Plummer-softened stream in the impulse approximation; allows for arbitrary velocity vectors, but y is input as the position along the stream

    INPUT:

       v - velocity of the stream (nstar,3)

       y - position along the stream (nstar)

       b - impact parameter

       w - velocity of the Plummer sphere (3)

       GSigma - surface density of the Plummer-softened stream (in natural units); should be a function of time

       rs - size of the Plummer sphere

       tmin, tmax= (None) minimum and maximum time to consider for GSigma (need to be set)

    OUTPUT:

       deltav (nstar,3)

    HISTORY:

       2015-11-14 - Written - Bovy (UofT)

    """
    if len(v.shape) == 1: 
        v= numpy.reshape(v,(1,3))
        y= numpy.reshape(y,(1,1))
    if tmax is None or tmax is None:
        raise ValueError("tmin= and tmax= need to be set")
    nv= v.shape[0]
    vmag= numpy.sqrt(numpy.sum(v**2.,axis=1))
    # Build the rotation matrices and their inverse
    rot= _rotation_vy(v)
    rotinv= _rotation_vy(v,inv=True)
    # Rotate the perturbing stream's velocity to the stream frames
    tilew= numpy.sum(rot*numpy.tile(w,(nv,3,1)),axis=-1)
    # Use similar expressions to Denis'
    wperp= numpy.sqrt(tilew[:,0]**2.+tilew[:,2]**2.)
    wpar= numpy.sqrt(numpy.sum(v**2.,axis=1))-tilew[:,1]
    wmag2= wpar**2.+wperp**2.
    wmag= numpy.sqrt(wmag2)
    b2= b**2.
    rs2= rs**2.
    wperp2= wperp**2.
    out= numpy.empty_like(v)
    out[:,0]= [1./wmag[ii]\
                   *integrate.quad(_astream_integrand_x,
                                   tmin,tmax,args=(y[ii],
                                                    vmag[ii],b,tilew[ii],
                                                    b2,wmag2[ii],
                                                    wperp[ii],wperp2[ii],
                                                    wpar[ii],
                                                    GSigma,rs2))[0]
               for ii in range(len(y))]
    out[:,1]= [-wperp2[ii]/wmag[ii]\
                    *integrate.quad(_astream_integrand_y,
                                    tmin,tmax,args=(y[ii],
                                                    vmag[ii],
                                                    b2,wmag2[ii],
                                                    wperp2[ii],
                                                    GSigma,rs2))[0]
                for ii in range(len(y))]
    out[:,2]= [1./wmag[ii]\
                    *integrate.quad(_astream_integrand_z       ,
                                    tmin,tmax,args=(y[ii],
                                                    vmag[ii],b,tilew[ii],
                                                    b2,wmag2[ii],
                                                    wperp[ii],wperp2[ii],
                                                    wpar[ii],
                                                    GSigma,rs2))[0]
               for ii in range(len(y))]
    # Rotate back to the original frame
    return 2.0*numpy.sum(\
        rotinv*numpy.swapaxes(numpy.tile(out.T,(3,1,1)).T,1,2),axis=-1)

def _astream_integrand(t,b_,orb,tx,w,GSigma,rs2,tmin,compt):
    teval= tx-tmin-t
    b__= b_+numpy.array([orb.x(teval),orb.y(teval),orb.z(teval)])
    w = w-numpy.array([orb.vx(teval),orb.vy(teval),orb.vz(teval)])
    wmag = numpy.sqrt(numpy.sum(w**2))
    bdotw=numpy.sum(b__*w)/wmag
    denom= wmag*(numpy.sum(b__**2)+rs2-bdotw**2)
    denom = 1./denom
    return -2.0*GSigma(t)*(((b__.T-bdotw*w.T/wmag)*denom).T)[compt]

def _astream_integrate(b_,orb,tx,w,GSigma,rs2,otmin,tmin,tmax):
    return numpy.array([integrate.quad(_astream_integrand,tmin,tmax,
                                       args=(b_,orb,tx,w,GSigma,rs2,
                                             otmin,i))[0] \
                            for i in range(3)])

def impulse_deltav_plummerstream_curvedstream(v,x,t,b,w,x0,v0,GSigma,rs,
                                              galpot,tmin=None,tmax=None):
    """
    NAME:

       impulse_deltav_plummerstream_curvedstream

    PURPOSE:

       calculate the delta velocity to due an encounter with a Plummer sphere in the impulse approximation; allows for arbitrary velocity vectors, and arbitrary position along the stream; velocities and positions are assumed to lie along an orbit

    INPUT:

       v - velocity of the stream (nstar,3)

       x - position along the stream (nstar,3)

       t - times at which (v,x) are reached, wrt the closest impact t=0 (nstar)

       b - impact parameter

       w - velocity of the Plummer sphere (3)

       x0 - point of closest approach

       v0 - velocity of point of closest approach

       GSigma - surface density of the Plummer-softened stream (in natural units); should be a function of time

       rs - size of the Plummer sphere

       galpot - galpy Potential object or list thereof

       tmin, tmax= (None) minimum and maximum time to consider for GSigma

    OUTPUT:

       deltav (nstar,3)

    HISTORY:

       2015-11-20 - Written based on Plummer sphere above - Bovy (UofT)

    """
    galpot= flatten_potential(galpot)
    if len(v.shape) == 1: v= numpy.reshape(v,(1,3))
    if len(x.shape) == 1: x= numpy.reshape(x,(1,3))
    # Integrate an orbit to use to figure out where each (v,x) is at each time
    R, phi, z= bovy_coords.rect_to_cyl(x0[0],x0[1],x0[2])
    vR, vT, vz= bovy_coords.rect_to_cyl_vec(v0[0],v0[1],v0[2],R,phi,z,cyl=True)
    # First back, then forward to cover the entire range with 1 orbit
    o= Orbit([R,vR,vT,z,vz,phi]).flip()
    ts= numpy.linspace(0.,numpy.fabs(numpy.amin(t)+tmin),101)
    o.integrate(ts,galpot)
    o= o(ts[-1]).flip()
    ts= numpy.linspace(0.,numpy.amax(t)+tmax-numpy.amin(t)-tmin,201)
    o.integrate(ts,galpot)
    # Calculate kicks
    b0 = numpy.cross(w,v0)
    b0 *= b/numpy.sqrt(numpy.sum(b0**2))
    return numpy.array(list(map(lambda i:_astream_integrate(\
                    b0-x0,o,i,w,GSigma,rs**2.,numpy.amin(t)+tmin,
                    tmin,tmax),t)))

def _rotation_vy(v,inv=False):
    return _rotate_to_arbitrary_vector(v,[0,1,0],inv)



##############################################################################
#
#   bovy_plot.py: general wrappers for matplotlib plotting
#
#       'public' methods:
#                         bovy_end_print
#                         bovy_dens2d
#                         bovy_hist
#                         bovy_plot
#                         bovy_print
#                         scatterplot (like hogg_scatterplot)
#                         bovy_text
#
#                         this module also defines a custom matplotlib 
#                         projection in which the polar azimuth increases
#                         clockwise (as in, the Galaxy viewed from the NGP)
#                         
#############################################################################
#############################################################################
#Copyright (c) 2010 - 2013, Jo Bovy
#All rights reserved.
#
#Redistribution and use in source and binary forms, with or without 
#modification, are permitted provided that the following conditions are met:
#
#   Redistributions of source code must retain the above copyright notice, 
#      this list of conditions and the following disclaimer.
#   Redistributions in binary form must reproduce the above copyright notice, 
#      this list of conditions and the following disclaimer in the 
#      documentation and/or other materials provided with the distribution.
#   The name of the author may not be used to endorse or promote products 
#      derived from this software without specific prior written permission.
#
#THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
#INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
#BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
#OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
#AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
#WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#POSSIBILITY OF SUCH DAMAGE.
#############################################################################
import re
import math as m
import scipy as sc
from scipy import special
from scipy import interpolate
from scipy import ndimage
import matplotlib
import matplotlib.pyplot as pyplot
import matplotlib.ticker as ticker
import matplotlib.cm as cm
from matplotlib import rc
from matplotlib.ticker import NullFormatter
from matplotlib.projections import PolarAxes, register_projection
from matplotlib.transforms import Affine2D, Bbox, IdentityTransform
from mpl_toolkits.mplot3d import Axes3D
from galpy.util.config import __config__
if __config__.getboolean('plot','seaborn-bovy-defaults'):
    try:
        import seaborn as sns
    except: pass
    else:
        sns.set_style('ticks',
                      {'xtick.direction': u'in',
                       'ytick.direction': u'in',
                       'axes.labelsize': 18.0,
                       'axes.titlesize': 18.0,
                       'figure.figsize': sc.array([ 6.64,  4.  ]),
                       'grid.linewidth': 2.0,
                       'legend.fontsize': 18.0,
                       'lines.linewidth': 2.0,
                       'lines.markeredgewidth': 0.0,
                       'lines.markersize': 14.0,
                       'patch.linewidth': 0.6,
                       'xtick.labelsize': 16.0,
                       'xtick.major.pad': 14.0,
                       'xtick.major.width': 2.0,
                       'xtick.minor.width': 1.0,
                       'ytick.labelsize': 16.0,
                       'ytick.major.pad': 14.0,
                       'ytick.major.width': 2.0,})
_DEFAULTNCNTR= 10
def bovy_end_print(filename,**kwargs):
    """
    NAME:

       bovy_end_print

    PURPOSE:

       saves the current figure(s) to filename

    INPUT:

       filename - filename for plot (with extension)

    OPTIONAL INPUTS:

       format - file-format

    OUTPUT:

       (none)

    HISTORY:

       2009-12-23 - Written - Bovy (NYU)

    """
    if 'format' in kwargs:
        pyplot.savefig(filename,**kwargs)
    else:
        pyplot.savefig(filename,format=re.split(r'\.',filename)[-1],**kwargs)
    pyplot.close()

def bovy_hist(x,xlabel=None,ylabel=None,overplot=False,**kwargs):
    """
    NAME:

       bovy_hist

    PURPOSE:

       wrapper around matplotlib's hist function

    INPUT:

       x - array to histogram

       xlabel - (raw string!) x-axis label, LaTeX math mode, no $s needed

       ylabel - (raw string!) y-axis label, LaTeX math mode, no $s needed

       yrange - set the y-axis range

       +all pyplot.hist keywords

    OUTPUT:
       (from the matplotlib docs:
       http://matplotlib.sourceforge.net/api/pyplot_api.html#matplotlib.pyplot.hist)

       The return value is a tuple (n, bins, patches)
       or ([n0, n1, ...], bins, [patches0, patches1,...])
       if the input contains multiple data

    HISTORY:

       2009-12-23 - Written - Bovy (NYU)

    """
    if not overplot:
        pyplot.figure()
    if 'xrange' in kwargs:
        xlimits= kwargs.pop('xrange')
        if not 'range' in kwargs:
            kwargs['range']= xlimits
        xrangeSet= True
    else: xrangeSet= False
    if 'yrange' in kwargs:
        ylimits= kwargs.pop('yrange')
        yrangeSet= True
    else: yrangeSet= False
    out= pyplot.hist(x,**kwargs)
    if overplot: return out
    _add_axislabels(xlabel,ylabel)
    if not 'range' in kwargs and not xrangeSet:
        if isinstance(x,list):
            xlimits=(sc.array(x).min(),sc.array(x).max())
        else:
            pyplot.xlim(x.min(),x.max())
    elif xrangeSet:
        pyplot.xlim(xlimits)
    else:
        pyplot.xlim(kwargs['range'])
    if yrangeSet:
        pyplot.ylim(ylimits)
    _add_ticks()
    return out

def bovy_plot(*args,**kwargs):
    """
    NAME:

       bovy_plot

    PURPOSE:

       wrapper around matplotlib's plot function

    INPUT:

       see http://matplotlib.sourceforge.net/api/pyplot_api.html#matplotlib.pyplot.plot

       xlabel - (raw string!) x-axis label, LaTeX math mode, no $s needed

       ylabel - (raw string!) y-axis label, LaTeX math mode, no $s needed

       xrange

       yrange

       scatter= if True, use pyplot.scatter and its options etc.

       colorbar= if True, and scatter==True, add colorbar

       crange - range for colorbar of scatter==True

       clabel= label for colorbar

       overplot=True does not start a new figure and does not change the ranges and labels

       gcf=True does not start a new figure (does change the ranges and labels)

       onedhists - if True, make one-d histograms on the sides

       onedhistcolor, onedhistfc, onedhistec

       onedhistxnormed, onedhistynormed - normed keyword for one-d histograms
       
       onedhistxweights, onedhistyweights - weights keyword for one-d histograms

       bins= number of bins for onedhists

       semilogx=, semilogy=, loglog= if True, plot logs

    OUTPUT:

       plot to output device, returns what pyplot.plot returns, or 3 Axes instances if onedhists=True

    HISTORY:

       2009-12-28 - Written - Bovy (NYU)

    """
    overplot= kwargs.pop('overplot',False)
    gcf= kwargs.pop('gcf',False)
    onedhists= kwargs.pop('onedhists',False)
    scatter= kwargs.pop('scatter',False)
    loglog= kwargs.pop('loglog',False)
    semilogx= kwargs.pop('semilogx',False)
    semilogy= kwargs.pop('semilogy',False)
    colorbar= kwargs.pop('colorbar',False)
    onedhisttype= kwargs.pop('onedhisttype','step')
    onedhistcolor= kwargs.pop('onedhistcolor','k')
    onedhistfc= kwargs.pop('onedhistfc','w')
    onedhistec= kwargs.pop('onedhistec','k')
    onedhistxnormed= kwargs.pop('onedhistxnormed',True)
    onedhistynormed= kwargs.pop('onedhistynormed',True)
    onedhistxweights= kwargs.pop('onedhistxweights',None)
    onedhistyweights= kwargs.pop('onedhistyweights',None)
    if 'bins' in kwargs:
        bins= kwargs['bins']
        kwargs.pop('bins')
    elif onedhists:
        if isinstance(args[0],sc.ndarray):
            bins= round(0.3*sc.sqrt(args[0].shape[0]))
        elif isinstance(args[0],list):
            bins= round(0.3*sc.sqrt(len(args[0])))
        else:
            bins= 30
    if onedhists:
        if overplot or gcf: fig= pyplot.gcf()
        else: fig= pyplot.figure()
        nullfmt   = NullFormatter()         # no labels
        # definitions for the axes
        left, width = 0.1, 0.65
        bottom, height = 0.1, 0.65
        bottom_h = left_h = left+width
        rect_scatter = [left, bottom, width, height]
        rect_histx = [left, bottom_h, width, 0.2]
        rect_histy = [left_h, bottom, 0.2, height]
        axScatter = pyplot.axes(rect_scatter)
        axHistx = pyplot.axes(rect_histx)
        axHisty = pyplot.axes(rect_histy)
        # no labels
        axHistx.xaxis.set_major_formatter(nullfmt)
        axHistx.yaxis.set_major_formatter(nullfmt)
        axHisty.xaxis.set_major_formatter(nullfmt)
        axHisty.yaxis.set_major_formatter(nullfmt)
        fig.sca(axScatter)
    elif not overplot and not gcf: pyplot.figure()
    ax=pyplot.gca()
    ax.set_autoscale_on(False)
    xlabel= kwargs.pop('xlabel',None)
    ylabel= kwargs.pop('ylabel',None)
    clabel= kwargs.pop('clabel',None)
    xlimits= kwargs.pop('xrange',None)
    if xlimits is None:
        if isinstance(args[0],list):
            xlimits=(sc.array(args[0]).min(),sc.array(args[0]).max())
        else:
            xlimits=(args[0].min(),args[0].max())
    ylimits= kwargs.pop('yrange',None)
    if ylimits is None:
        if isinstance(args[1],list):
            ylimits=(sc.array(args[1]).min(),sc.array(args[1]).max())
        else:
            ylimits=(args[1].min(),args[1].max())
    climits= kwargs.pop('crange',None)
    if climits is None and scatter:
        if 'c' in kwargs and isinstance(kwargs['c'],list):
            climits=(sc.array(kwargs['c']).min(),sc.array(kwargs['c']).max())
        elif 'c' in kwargs:
            climits=(kwargs['c'].min(),kwargs['c'].max())
        else:
            climits= None
    if scatter:
        out= pyplot.scatter(*args,**kwargs)
    elif loglog:
        out= pyplot.loglog(*args,**kwargs)
    elif semilogx:
        out= pyplot.semilogx(*args,**kwargs)
    elif semilogy:
        out= pyplot.semilogy(*args,**kwargs)
    else:
        out= pyplot.plot(*args,**kwargs)
    if overplot:
        pass
    else:
        if semilogy:
            ax= pyplot.gca()
            ax.set_yscale('log')
        elif semilogx:
            ax= pyplot.gca()
            ax.set_xscale('log')
        elif loglog:
            ax= pyplot.gca()
            ax.set_xscale('log')
            ax.set_yscale('log')
        pyplot.xlim(*xlimits)
        pyplot.ylim(*ylimits)
        _add_axislabels(xlabel,ylabel)
        if not semilogy and not semilogx and not loglog:
            _add_ticks()
        elif semilogy:
            _add_ticks(xticks=True,yticks=False)
        elif semilogx:
            _add_ticks(yticks=True,xticks=False)
    #Add colorbar
    if colorbar:
        cbar= pyplot.colorbar(out,fraction=0.15)
        cbar.set_clim(*climits)
        if not clabel is None:
            cbar.set_label(clabel)
    #Add onedhists
    if not onedhists:
        return out
    histx, edges, patches= axHistx.hist(args[0], bins=bins,
                                        normed=onedhistxnormed,
                                        weights=onedhistxweights,
                                        histtype=onedhisttype,
                                        range=sorted(xlimits),
                                        color=onedhistcolor,fc=onedhistfc,
                                        ec=onedhistec)
    histy, edges, patches= axHisty.hist(args[1], bins=bins,
                                        orientation='horizontal',
                                        weights=onedhistyweights,
                                        normed=onedhistynormed,
                                        histtype=onedhisttype,
                                        range=sorted(ylimits),
                                        color=onedhistcolor,fc=onedhistfc,
                                        ec=onedhistec)
    axHistx.set_xlim( axScatter.get_xlim() )
    axHisty.set_ylim( axScatter.get_ylim() )
    axHistx.set_ylim( 0, 1.2*sc.amax(histx))
    axHisty.set_xlim( 0, 1.2*sc.amax(histy))
    return (axScatter,axHistx,axHisty)

def bovy_plot3d(*args,**kwargs):
    """
    NAME:

       bovy_plot3d

    PURPOSE:

       plot in 3d much as in 2d

    INPUT:

       see http://matplotlib.sourceforge.net/api/pyplot_api.html#matplotlib.pyplot.plot

       xlabel - (raw string!) x-axis label, LaTeX math mode, no $s needed

       ylabel - (raw string!) y-axis label, LaTeX math mode, no $s needed

       xrange

       yrange

       overplot=True does not start a new figure

    OUTPUT:

    HISTORY:

       2011-01-08 - Written - Bovy (NYU)

    """
    overplot= kwargs.pop('overplot',False)
    if not overplot: pyplot.figure()
    ax=pyplot.gca(projection='3d')
    ax.set_autoscale_on(False)
    xlabel= kwargs.pop('xlabel',None)
    ylabel= kwargs.pop('ylabel',None)
    zlabel= kwargs.pop('zlabel',None)
    if 'xrange' in kwargs:
        xlimits= kwargs.pop('xrange')
    else:
        if isinstance(args[0],list):
            xlimits=(sc.array(args[0]).min(),sc.array(args[0]).max())
        else:
            xlimits=(args[0].min(),args[0].max())
    if 'yrange' in kwargs:
        ylimits= kwargs.pop('yrange')
    else:
        if isinstance(args[1],list):
            ylimits=(sc.array(args[1]).min(),sc.array(args[1]).max())
        else:
            ylimits=(args[1].min(),args[1].max())
    if 'zrange' in kwargs:
        zlimits= kwargs.pop('zrange')
    else:
        if isinstance(args[2],list):
            zlimits=(sc.array(args[2]).min(),sc.array(args[2]).max())
        else:
            zlimits=(args[1].min(),args[2].max())
    out= pyplot.plot(*args,**kwargs)
    if overplot:
        pass
    else:
        if xlabel != None:
            if xlabel[0] != '$':
                thisxlabel=r'$'+xlabel+'$'
            else:
                thisxlabel=xlabel
            ax.set_xlabel(thisxlabel)
        if ylabel != None:
            if ylabel[0] != '$':
                thisylabel=r'$'+ylabel+'$'
            else:
                thisylabel=ylabel
            ax.set_ylabel(thisylabel)
        if zlabel != None:
            if zlabel[0] != '$':
                thiszlabel=r'$'+zlabel+'$'
            else:
                thiszlabel=zlabel
            ax.set_zlabel(thiszlabel)
        ax.set_xlim3d(*xlimits)
        ax.set_ylim3d(*ylimits)
        ax.set_zlim3d(*zlimits)
    return out

def bovy_dens2d(X,**kwargs):
    """
    NAME:

       bovy_dens2d

    PURPOSE:

       plot a 2d density with optional contours

    INPUT:

       first argument is the density

       matplotlib.pyplot.imshow keywords (see http://matplotlib.sourceforge.net/api/axes_api.html#matplotlib.axes.Axes.imshow)

       xlabel - (raw string!) x-axis label, LaTeX math mode, no $s needed

       ylabel - (raw string!) y-axis label, LaTeX math mode, no $s needed

       xrange

       yrange

       noaxes - don't plot any axes

       overplot - if True, overplot

       colorbar - if True, add colorbar

       shrink= colorbar argument: shrink the colorbar by the factor (optional)

       conditional - normalize each column separately (for probability densities, i.e., cntrmass=True)

       gcf=True does not start a new figure (does change the ranges and labels)

       Contours:
       
       justcontours - if True, only draw contours

       contours - if True, draw contours (10 by default)

       levels - contour-levels

       cntrmass - if True, the density is a probability and the levels are probability masses contained within the contour

       cntrcolors - colors for contours (single color or array)

       cntrlabel - label the contours

       cntrlw, cntrls - linewidths and linestyles for contour

       cntrlabelsize, cntrlabelcolors,cntrinline - contour arguments

       cntrSmooth - use ndimage.gaussian_filter to smooth before contouring

       onedhists - if True, make one-d histograms on the sides

       onedhistcolor - histogram color

       retAxes= return all Axes instances

       retCont= return the contour instance

    OUTPUT:

       plot to output device, Axes instances depending on input

    HISTORY:

       2010-03-09 - Written - Bovy (NYU)

    """
    overplot= kwargs.pop('overplot',False)
    gcf= kwargs.pop('gcf',False)
    if not overplot and not gcf:
        pyplot.figure()
    xlabel= kwargs.pop('xlabel',None)
    ylabel= kwargs.pop('ylabel',None)
    zlabel= kwargs.pop('zlabel',None)
    if 'extent' in kwargs:
        extent= kwargs.pop('extent')
    else:
        xlimits= kwargs.pop('xrange',[0,X.shape[1]])
        ylimits= kwargs.pop('yrange',[0,X.shape[0]])
        extent= xlimits+ylimits
    if not 'aspect' in kwargs:
        kwargs['aspect']= (xlimits[1]-xlimits[0])/float(ylimits[1]-ylimits[0])
    noaxes= kwargs.pop('noaxes',False)
    justcontours= kwargs.pop('justcontours',False)
    if ('contours' in kwargs and kwargs['contours']) or \
            'levels' in kwargs or justcontours or \
            ('cntrmass' in kwargs and kwargs['cntrmass']):
        contours= True
    else:
        contours= False
    kwargs.pop('contours',None)
    if 'levels' in kwargs:
        levels= kwargs['levels']
        kwargs.pop('levels')
    elif contours:
        if 'cntrmass' in kwargs and kwargs['cntrmass']:
            levels= sc.linspace(0.,1.,_DEFAULTNCNTR)
        elif True in sc.isnan(sc.array(X)):
            levels= sc.linspace(sc.nanmin(X),sc.nanmax(X),_DEFAULTNCNTR)
        else:
            levels= sc.linspace(sc.amin(X),sc.amax(X),_DEFAULTNCNTR)
    cntrmass= kwargs.pop('cntrmass',False)
    conditional= kwargs.pop('conditional',False)
    cntrcolors= kwargs.pop('cntrcolors','k')
    cntrlabel= kwargs.pop('cntrlabel',False)
    cntrlw= kwargs.pop('cntrlw',None)
    cntrls= kwargs.pop('cntrls',None)
    cntrSmooth= kwargs.pop('cntrSmooth',None)
    cntrlabelsize= kwargs.pop('cntrlabelsize',None)
    cntrlabelcolors= kwargs.pop('cntrlabelcolors',None)
    cntrinline= kwargs.pop('cntrinline',None)
    retCumImage= kwargs.pop('retCumImage',False)
    cb= kwargs.pop('colorbar',False)
    shrink= kwargs.pop('shrink',None)
    onedhists= kwargs.pop('onedhists',False)
    onedhistcolor= kwargs.pop('onedhistcolor','k')
    retAxes= kwargs.pop('retAxes',False)
    retCont= kwargs.pop('retCont',False)
    if onedhists:
        if overplot or gcf: fig= pyplot.gcf()
        else: fig= pyplot.figure()
        nullfmt   = NullFormatter()         # no labels
        # definitions for the axes
        left, width = 0.1, 0.65
        bottom, height = 0.1, 0.65
        bottom_h = left_h = left+width
        rect_scatter = [left, bottom, width, height]
        rect_histx = [left, bottom_h, width, 0.2]
        rect_histy = [left_h, bottom, 0.2, height]
        axScatter = pyplot.axes(rect_scatter)
        axHistx = pyplot.axes(rect_histx)
        axHisty = pyplot.axes(rect_histy)
        # no labels
        axHistx.xaxis.set_major_formatter(nullfmt)
        axHistx.yaxis.set_major_formatter(nullfmt)
        axHisty.xaxis.set_major_formatter(nullfmt)
        axHisty.yaxis.set_major_formatter(nullfmt)
        fig.sca(axScatter)
    ax=pyplot.gca()
    ax.set_autoscale_on(False)
    if conditional:
        plotthis= X/sc.tile(sc.sum(X,axis=0),(X.shape[1],1))
    else:
        plotthis= X
    if not justcontours:
        out= pyplot.imshow(plotthis,extent=extent,**kwargs)
    if not overplot:
        pyplot.axis(extent)
        _add_axislabels(xlabel,ylabel)
        _add_ticks()
    #Add colorbar
    if cb and not justcontours:
        if shrink is None:
            shrink= sc.amin([float(kwargs.pop('aspect',1.))*0.87,1.])
        CB1= pyplot.colorbar(out,shrink=shrink)
        if not zlabel is None:
            if zlabel[0] != '$':
                thiszlabel=r'$'+zlabel+'$'
            else:
                thiszlabel=zlabel
            CB1.set_label(thiszlabel)
    if contours or retCumImage:
        aspect= kwargs.get('aspect',None)
        origin= kwargs.get('origin',None)
        if cntrmass:
            #Sum from the top down!
            plotthis[sc.isnan(plotthis)]= 0.
            sortindx= sc.argsort(plotthis.flatten())[::-1]
            cumul= sc.cumsum(sc.sort(plotthis.flatten())[::-1])/sc.sum(plotthis.flatten())
            cntrThis= sc.zeros(sc.prod(plotthis.shape))
            cntrThis[sortindx]= cumul
            cntrThis= sc.reshape(cntrThis,plotthis.shape)
        else:
            cntrThis= plotthis
        if contours:
            if not cntrSmooth is None:
                cntrThis= ndimage.gaussian_filter(cntrThis,cntrSmooth,
                                                  mode='nearest')
            cont= pyplot.contour(cntrThis,levels,colors=cntrcolors,
                                 linewidths=cntrlw,extent=extent,aspect=aspect,
                                 linestyles=cntrls,origin=origin)
            if cntrlabel:
                pyplot.clabel(cont,fontsize=cntrlabelsize,
                              colors=cntrlabelcolors,
                              inline=cntrinline)
    if noaxes:
        ax.set_axis_off()
    #Add onedhists
    if not onedhists:
        if retCumImage:
            return cntrThis
        elif retAxes:
            return pyplot.gca()
        elif retCont:
            return cont
        elif justcontours:
            return cntrThis
        else:
            return out
    histx= sc.nansum(X.T,axis=1)*m.fabs(ylimits[1]-ylimits[0])/X.shape[1] #nansum bc nan is *no dens value*
    histy= sc.nansum(X.T,axis=0)*m.fabs(xlimits[1]-xlimits[0])/X.shape[0]
    histx[sc.isnan(histx)]= 0.
    histy[sc.isnan(histy)]= 0.
    dx= (extent[1]-extent[0])/float(len(histx))
    axHistx.plot(sc.linspace(extent[0]+dx,extent[1]-dx,len(histx)),histx,
                 drawstyle='steps-mid',color=onedhistcolor)
    dy= (extent[3]-extent[2])/float(len(histy))
    axHisty.plot(histy,sc.linspace(extent[2]+dy,extent[3]-dy,len(histy)),
                 drawstyle='steps-mid',color=onedhistcolor)
    axHistx.set_xlim( axScatter.get_xlim() )
    axHisty.set_ylim( axScatter.get_ylim() )
    axHistx.set_ylim( 0, 1.2*sc.amax(histx))
    axHisty.set_xlim( 0, 1.2*sc.amax(histy))
    if retCumImage:
        return cntrThis
    elif retAxes:
        return (axScatter,axHistx,axHisty)
    elif justcontours:
        return cntrThis
    else:
        return out

def bovy_print(fig_width=5,fig_height=5,axes_labelsize=16,
               text_fontsize=11,legend_fontsize=12,
               xtick_labelsize=10,ytick_labelsize=10,
               xtick_minor_size=2,ytick_minor_size=2,
               xtick_major_size=4,ytick_major_size=4):
    """
    NAME:

       bovy_print

    PURPOSE:

       setup a figure for plotting

    INPUT:

       fig_width - width in inches

       fig_height - height in inches

       axes_labelsize - size of the axis-labels

       text_fontsize - font-size of the text (if any)

       legend_fontsize - font-size of the legend (if any)

       xtick_labelsize - size of the x-axis labels

       ytick_labelsize - size of the y-axis labels

       xtick_minor_size - size of the minor x-ticks

       ytick_minor_size - size of the minor y-ticks

    OUTPUT:

       (none)

    HISTORY:

       2009-12-23 - Written - Bovy (NYU)

    """
    fig_size =  [fig_width,fig_height]
    params = {'axes.labelsize': axes_labelsize,
              'font.size': text_fontsize,
              'legend.fontsize': legend_fontsize,
              'xtick.labelsize':xtick_labelsize,
              'ytick.labelsize':ytick_labelsize,
              'text.usetex': True,
              'figure.figsize': fig_size,
              'xtick.major.size' : xtick_major_size,
              'ytick.major.size' : ytick_major_size,
              'xtick.minor.size' : xtick_minor_size,
              'ytick.minor.size' : ytick_minor_size,
              'legend.numpoints':1,
              'xtick.top': True,
              'xtick.direction': 'in',
              'ytick.right': True,
              'ytick.direction': 'in'}
    pyplot.rcParams.update(params)
    rc('text.latex', preamble=r'\usepackage{amsmath}'+'\n'
       +r'\usepackage{amssymb}')

def bovy_text(*args,**kwargs):
    """
    NAME:

       bovy_text

    PURPOSE:

       thin wrapper around matplotlib's text and annotate

       use keywords:

          'bottom_left=True'

          'bottom_right=True'

          'top_left=True'

          'top_right=True'

          'title=True'

       to place the text in one of the corners or use it as the title

    INPUT:

       see matplotlib's text
          (http://matplotlib.sourceforge.net/api/pyplot_api.html#matplotlib.pyplot.text)

    OUTPUT:

       prints text on the current figure

    HISTORY:

       2010-01-26 - Written - Bovy (NYU)

    """
    if kwargs.pop('title',False):
        pyplot.annotate(args[0],(0.5,1.05),xycoords='axes fraction',
                        horizontalalignment='center',
                        verticalalignment='top',**kwargs)
    elif kwargs.pop('bottom_left',False):
        pyplot.annotate(args[0],(0.05,0.05),xycoords='axes fraction',**kwargs)
    elif kwargs.pop('bottom_right',False):
        pyplot.annotate(args[0],(0.95,0.05),xycoords='axes fraction',
                        horizontalalignment='right',**kwargs)
    elif kwargs.pop('top_right',False):
        pyplot.annotate(args[0],(0.95,0.95),xycoords='axes fraction',
                        horizontalalignment='right',
                        verticalalignment='top',**kwargs)
    elif kwargs.pop('top_left',False):
        pyplot.annotate(args[0],(0.05,0.95),xycoords='axes fraction',
                        verticalalignment='top',**kwargs)
    else:
        pyplot.text(*args,**kwargs)

def scatterplot(x,y,*args,**kwargs):
    """
    NAME:

       scatterplot

    PURPOSE:

       make a 'smart' scatterplot that is a density plot in high-density
       regions and a regular scatterplot for outliers

    INPUT:

       x, y

       xlabel - (raw string!) x-axis label, LaTeX math mode, no $s needed

       ylabel - (raw string!) y-axis label, LaTeX math mode, no $s needed

       xrange

       yrange

       bins - number of bins to use in each dimension

       weights - data-weights

       aspect - aspect ratio

       conditional - normalize each column separately (for probability densities, i.e., cntrmass=True)

       gcf=True does not start a new figure (does change the ranges and labels)

       contours - if False, don't plot contours

       justcontours - if True, only draw contours, no density

       cntrcolors - color of contours (can be array as for bovy_dens2d)

       cntrlw, cntrls - linewidths and linestyles for contour

       cntrSmooth - use ndimage.gaussian_filter to smooth before contouring

       levels - contour-levels; data points outside of the last level will be individually shown (so, e.g., if this list is descending, contours and data points will be overplotted)

       onedhists - if True, make one-d histograms on the sides

       onedhistx - if True, make one-d histograms on the side of the x distribution

       onedhisty - if True, make one-d histograms on the side of the y distribution

       onedhistcolor, onedhistfc, onedhistec

       onedhistxnormed, onedhistynormed - normed keyword for one-d histograms
       
       onedhistxweights, onedhistyweights - weights keyword for one-d histograms

       cmap= cmap for density plot

       hist= and edges= - you can supply the histogram of the data yourself, this can be useful if you want to censor the data, both need to be set and calculated using scipy.histogramdd with the given range

       retAxes= return all Axes instances

    OUTPUT:

       plot to output device, Axes instance(s) or not, depending on input

    HISTORY:

       2010-04-15 - Written - Bovy (NYU)

    """
    xlabel= kwargs.pop('xlabel',None)
    ylabel= kwargs.pop('ylabel',None)
    if 'xrange' in kwargs:
        xrange= kwargs.pop('xrange')
    else:
        if isinstance(x,list): xrange=[sc.amin(x),sc.amax(x)]
        else: xrange=[x.min(),x.max()]
    if 'yrange' in kwargs:
        yrange= kwargs.pop('yrange')
    else:
        if isinstance(y,list): yrange=[sc.amin(y),sc.amax(y)]
        else: yrange=[y.min(),y.max()]
    ndata= len(x)
    bins= kwargs.pop('bins',round(0.3*sc.sqrt(ndata)))
    weights= kwargs.pop('weights',None)
    levels= kwargs.pop('levels',special.erf(sc.arange(1,4)/sc.sqrt(2.)))
    aspect= kwargs.pop('aspect',(xrange[1]-xrange[0])/(yrange[1]-yrange[0]))
    conditional= kwargs.pop('conditional',False)
    contours= kwargs.pop('contours',True)
    justcontours= kwargs.pop('justcontours',False)
    cntrcolors= kwargs.pop('cntrcolors','k')
    cntrlw= kwargs.pop('cntrlw',None)
    cntrls= kwargs.pop('cntrls',None)
    cntrSmooth= kwargs.pop('cntrSmooth',None)
    onedhists= kwargs.pop('onedhists',False)
    onedhistx= kwargs.pop('onedhistx',onedhists)
    onedhisty= kwargs.pop('onedhisty',onedhists)
    onedhisttype= kwargs.pop('onedhisttype','step')
    onedhistcolor= kwargs.pop('onedhistcolor','k')
    onedhistfc= kwargs.pop('onedhistfc','w')
    onedhistec= kwargs.pop('onedhistec','k')
    onedhistls= kwargs.pop('onedhistls','solid')
    onedhistlw= kwargs.pop('onedhistlw',None)
    onedhistsbins= kwargs.pop('onedhistsbins',round(0.3*sc.sqrt(ndata)))
    overplot= kwargs.pop('overplot',False)
    gcf= kwargs.pop('gcf',False)
    cmap= kwargs.pop('cmap',cm.gist_yarg)
    onedhistxnormed= kwargs.pop('onedhistxnormed',True)
    onedhistynormed= kwargs.pop('onedhistynormed',True)
    onedhistxweights= kwargs.pop('onedhistxweights',weights)
    onedhistyweights= kwargs.pop('onedhistyweights',weights)
    retAxes= kwargs.pop('retAxes',False)
    if onedhists or onedhistx or onedhisty:
        if overplot or gcf: fig= pyplot.gcf()
        else: fig= pyplot.figure()
        nullfmt   = NullFormatter()         # no labels
        # definitions for the axes
        left, width = 0.1, 0.65
        bottom, height = 0.1, 0.65
        bottom_h = left_h = left+width
        rect_scatter = [left, bottom, width, height]
        rect_histx = [left, bottom_h, width, 0.2]
        rect_histy = [left_h, bottom, 0.2, height]
        axScatter = pyplot.axes(rect_scatter)
        if onedhistx:
            axHistx = pyplot.axes(rect_histx)
            # no labels
            axHistx.xaxis.set_major_formatter(nullfmt)
            axHistx.yaxis.set_major_formatter(nullfmt)
        if onedhisty:
            axHisty = pyplot.axes(rect_histy)
            # no labels
            axHisty.xaxis.set_major_formatter(nullfmt)
            axHisty.yaxis.set_major_formatter(nullfmt)
        fig.sca(axScatter)
    data= sc.array([x,y]).T
    if 'hist' in kwargs and 'edges' in kwargs:
        hist=kwargs['hist']
        kwargs.pop('hist')
        edges=kwargs['edges']
        kwargs.pop('edges')
    else:
        hist, edges= sc.histogramdd(data,bins=bins,range=[xrange,yrange],
                                    weights=weights)
    if contours:
        cumimage= bovy_dens2d(hist.T,contours=contours,levels=levels,
                              cntrmass=contours,cntrSmooth=cntrSmooth,
                              cntrcolors=cntrcolors,cmap=cmap,origin='lower',
                              xrange=xrange,yrange=yrange,xlabel=xlabel,
                              ylabel=ylabel,interpolation='nearest',
                              retCumImage=True,aspect=aspect,
                              conditional=conditional,
                              cntrlw=cntrlw,cntrls=cntrls,
                              justcontours=justcontours,zorder=5*justcontours,
                              overplot=(gcf or onedhists or overplot or onedhistx or onedhisty))
    else:
        cumimage= bovy_dens2d(hist.T,contours=contours,
                              cntrcolors=cntrcolors,
                              cmap=cmap,origin='lower',
                              xrange=xrange,yrange=yrange,xlabel=xlabel,
                              ylabel=ylabel,interpolation='nearest',
                              conditional=conditional,
                              retCumImage=True,aspect=aspect,
                              cntrlw=cntrlw,cntrls=cntrls,
                              overplot=(gcf or onedhists or overplot or onedhistx or onedhisty))
    #Set axes and labels
    pyplot.axis(list(xrange)+list(yrange))
    if not overplot:
        _add_axislabels(xlabel,ylabel)
        _add_ticks()
    binxs= []
    xedge= edges[0]
    for ii in range(len(xedge)-1):
        binxs.append((xedge[ii]+xedge[ii+1])/2.)
    binxs= sc.array(binxs)
    binys= []
    yedge= edges[1]
    for ii in range(len(yedge)-1):
        binys.append((yedge[ii]+yedge[ii+1])/2.)
    binys= sc.array(binys)
    cumInterp= interpolate.RectBivariateSpline(binxs,binys,cumimage.T,
                                               kx=1,ky=1)
    cums= []
    for ii in range(len(x)):
        cums.append(cumInterp(x[ii],y[ii])[0,0])
    cums= sc.array(cums)
    plotx= x[cums > levels[-1]]
    ploty= y[cums > levels[-1]]
    if not len(plotx) == 0:
        if not weights == None:
            w8= weights[cums > levels[-1]]
            for ii in range(len(plotx)):
                bovy_plot(plotx[ii],ploty[ii],overplot=True,
                          color='%.2f'%(1.-w8[ii]),*args,**kwargs)
        else:
            bovy_plot(plotx,ploty,overplot=True,zorder=1,*args,**kwargs)
    #Add onedhists
    if not (onedhists or onedhistx or onedhisty):
        if retAxes:
            return pyplot.gca()
        else:
            return None
    if onedhistx:
        histx, edges, patches= axHistx.hist(x,bins=onedhistsbins,
                                            normed=onedhistxnormed,
                                            weights=onedhistxweights,
                                            histtype=onedhisttype,
                                            range=sorted(xrange),
                                            color=onedhistcolor,fc=onedhistfc,
                                            ec=onedhistec,ls=onedhistls,
                                            lw=onedhistlw)
    if onedhisty:
        histy, edges, patches= axHisty.hist(y,bins=onedhistsbins,
                                            orientation='horizontal',
                                            weights=onedhistyweights,
                                            normed=onedhistynormed,
                                            histtype=onedhisttype,
                                            range=sorted(yrange),
                                            color=onedhistcolor,fc=onedhistfc,
                                            ec=onedhistec,ls=onedhistls,
                                            lw=onedhistlw)
    if onedhistx and not overplot:
        axHistx.set_xlim( axScatter.get_xlim() )
        axHistx.set_ylim( 0, 1.2*sc.amax(histx))
    if onedhisty and not overplot:
        axHisty.set_ylim( axScatter.get_ylim() )
        axHisty.set_xlim( 0, 1.2*sc.amax(histy))
    if not onedhistx: axHistx= None
    if not onedhisty: axHisty= None
    if retAxes:
        return (axScatter,axHistx,axHisty)
    else:
        return None

def _add_axislabels(xlabel,ylabel):
    """
    NAME:

       _add_axislabels

    PURPOSE:

       add axis labels to the current figure

    INPUT:

       xlabel - (raw string!) x-axis label, LaTeX math mode, no $s needed

       ylabel - (raw string!) y-axis label, LaTeX math mode, no $s needed

    OUTPUT:

       (none; works on the current axes)

    HISTORY:

       2009-12-23 - Written - Bovy (NYU)

    """
    if xlabel != None:
        if xlabel[0] != '$':
            thisxlabel=r'$'+xlabel+'$'
        else:
            thisxlabel=xlabel
        pyplot.xlabel(thisxlabel)
    if ylabel != None:
        if ylabel[0] != '$':
            thisylabel=r'$'+ylabel+'$'
        else:
            thisylabel=ylabel
        pyplot.ylabel(thisylabel)
        
def _add_ticks(xticks=True,yticks=True):
    """
    NAME:

       _add_ticks

    PURPOSE:

       add minor axis ticks to a plot

    INPUT:

       (none; works on the current axes)

    OUTPUT:

       (none; works on the current axes)

    HISTORY:

       2009-12-23 - Written - Bovy (NYU)

    """
    ax=pyplot.gca()
    if xticks:
        xstep= ax.xaxis.get_majorticklocs()
        xstep= xstep[1]-xstep[0]
        ax.xaxis.set_minor_locator(ticker.MultipleLocator(xstep/5.))
    if yticks:
        ystep= ax.yaxis.get_majorticklocs()
        ystep= ystep[1]-ystep[0]
        ax.yaxis.set_minor_locator(ticker.MultipleLocator(ystep/5.))


class GalPolarAxes(PolarAxes):
    '''
    A variant of PolarAxes where theta increases clockwise
    '''
    name = 'galpolar'

    class GalPolarTransform(PolarAxes.PolarTransform):
        def transform(self, tr):
            xy   = sc.zeros(tr.shape, sc.float_)
            t    = tr[:, 0:1]
            r    = tr[:, 1:2]
            x    = xy[:, 0:1]
            y    = xy[:, 1:2]
            x[:] = r * sc.cos(t)
            y[:] = -r * sc.sin(t)
            return xy

        transform_non_affine = transform

        def inverted(self):
            return GalPolarAxes.InvertedGalPolarTransform()

    class InvertedGalPolarTransform(PolarAxes.InvertedPolarTransform):
        def transform(self, xy):
            x = xy[:, 0:1]
            y = xy[:, 1:]
            r = sc.sqrt(x*x + y*y)
            theta = sc.arctan2(y, x)
            return sc.concatenate((theta, r), 1)

        def inverted(self):
            return GalPolarAxes.GalPolarTransform()

    def _set_lim_and_transforms(self):
        PolarAxes._set_lim_and_transforms(self)
        self.transProjection = self.GalPolarTransform()
        self.transData = (
            self.transScale + 
            self.transProjection + 
            (self.transProjectionAffine + self.transAxes))
        self._xaxis_transform = (
            self.transProjection +
            self.PolarAffine(IdentityTransform(), Bbox.unit()) +
            self.transAxes)
        self._xaxis_text1_transform = (
            self._theta_label1_position +
            self._xaxis_transform)
        self._yaxis_transform = (
            Affine2D().scale(sc.pi * 2.0, 1.0) +
            self.transData)
        self._yaxis_text1_transform = (
            self._r_label1_position +
            Affine2D().scale(1.0 / 360.0, 1.0) +
            self._yaxis_transform)

register_projection(GalPolarAxes)    
